import sys

from mock import patch

# TODO: create a more exhaustive list and move to a different module.
DEFAULT_EXTERNAL_MODULES = [
    # Django
    'django',
    'django.apps', 'django.test.client',
    'django.conf', 'django.conf.settings', 'django.core', 'django.core.exceptions',
    'django.contrib', 'django.contrib.auth', 'django.contrib.auth.decorators',
    'django.db', 'django.db.models',
    'django.http',
    'django.shortcuts',
    'django.test',
    'django.urls', 'django.urls.converters',
    'django.utils', 'django.utils.translation',
    'django.views', 'django.views.decorators',
    'django.views.decorators.cache', 'django.views.decorators.clickjacking',
    'django.views.decorators.csrf', 'django.views.decorators.gzip',

    # Rest Framework
    'rest_framework',
    'rest_framework.decorators',
    'rest_framework.relations', 'rest_framework.response',
    'rest_framework.serializers',
    'rest_framework.utils', 'rest_framework.utils.encoders',
    'rest_framework.viewsets',

    # Scientific
    'numpy', 'pandas',

    # Tools
    'pytest', 'faker', 'model_mommy', 'model_mommy.mommy',
    'factory', 'factory.declarations',
]


class LightMock:
    """
    A lightweight Mock class.

    It creates attributes and methods on-demand.

    >>> x = LightMock()
    >>> x.foo.bar(42)                                      # doctests: +ELLIPSIS
    <LightMock ...>
    """

    def __init__(self, *args, **kwargs):
        pass

    __method = (lambda self, *args, **kwargs: LightMock())
    __call__ = __method
    __getattr__ = __method
    ___name__ = ___qualname__ = 'mock'

    # Container interface
    __getitem__ = __method
    __iter__ = (lambda self: iter(()))

    # Arithmetic operations
    __add__ = __sub__ = __mul__ = __truediv__ = __floordiv__ = __method
    __radd__ = __rsub__ = __rmul__ = __rtruediv__ = __rfloordiv__ = __method

    # Conversions
    __bool__ = (lambda self: True)

    # Pretend to be a type in some situations
    __mro_entries__ = (lambda *args: (object,))
    __getstate__ = (lambda self: ())


def mock_save(model, method=LightMock):
    """
    Context manager that mocks the .save() method of a model to prevent it from
    hitting the database.

    Usage:
        .. code-block:: python

            with mock_save(model):
                model.name = "Hello"
                model.save()  # it does not actually touch the db
    """
    return patch.object(model, 'save', LightMock)


_stdout = sys.stdout


def mock_modules(*modules):
    """
    Save all given modules that were not imported in sys.modules.
    """

    cls = type('MockItem', (LightMock, type), {})

    if modules == ('auto',):
        modules = DEFAULT_EXTERNAL_MODULES

    def path_hook(path):
        base = path.partition('.')[0]
        if base in modules:
            # It mocks all its way through a mocked module In python import
            # subsystem. This hack makes the implementation much simpler :)
            return cls()
        else:
            raise ImportError

    sys.path_hooks.insert(0, path_hook)

    for mod in modules:
        sys.modules.setdefault(mod, LightMock())


def assume_unique(form=None):
    """
    Context manager that suppress checks of uniqueness during model validation.

    Usage:
        .. code-block:: python

            with assume_unique(model):
                model.slug = "repeated-slug"
                model.full_clean()  # prevents touching the db on uniqueness checks
    """
    from django import forms
    if form is None:
        form = forms.ModelForm
    return patch.object(form, 'validate_unique', no_op)


#
# Auxiliary functions
#
no_op = (lambda *args, **kwargs: None)
cte = (lambda cte: lambda *args, **kwargs: cte)


def raise_exception(exception):
    """
    Return a function that raises the given exception when called.
    """
    if exception is not None:
        def action(*args, **kwargs):
            raise exception
    else:
        action = no_op
    return action



import functools
from time import sleep

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from model_utils import Choices
from model_utils.models import StatusModel, TimeStampedModel

from boogie.apps.tasks.manager import TaskQuerySet
from boogie.apps.tasks.settings import REGISTERED_FUNCTIONS, USE_CELERY, log


class TaskMeta(type(models.Model)):
    """
    Metaclass that register @task-decorated methods as tasks.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self._meta.abstract:
            for attr in dir(self):
                method = getattr(self, attr, None)
                if getattr(method, 'is_task', False):
                    self.register(name=attr, function=method)


class Task(StatusModel, TimeStampedModel, metaclass=TaskMeta):
    """
    Represents a deferred computation.

    Concrete classes should override the object and result fields.
    """
    STATUS = Choices(
        ('pending', _('pending')),
        ('running', _('running')),
        ('finished', _('finished')),
        ('skipped', _('skipped')),
        ('failed', _('failed')),
    )
    function = models.CharField(max_length=40)
    arguments = JSONField(default=dict)
    object = models.ForeignKey('contenttypes.ContentType', on_delete=models.CASCADE)
    result = models.TextField(blank=True)

    is_pending = property(lambda self: self.status == self.STATUS.pending)
    is_running = property(lambda self: self.status == self.STATUS.running)
    is_finished = property(lambda self: self.status == self.STATUS.finished)
    is_skipped = property(lambda self: self.status == self.STATUS.skipped)
    is_failed = property(lambda self: self.status == self.STATUS.failed)

    objects = TaskQuerySet.as_manager()

    class Meta:
        abstract = True

    class Skip(Exception):
        """
        Exception raised when task should be skipped
        """

    @classmethod
    def get_object_model(cls):
        """
        Return model class for the object field.
        """
        field = cls._meta.get_field('object')
        return field.remote_field.model

    @classmethod
    def _create(cls, **kwargs):
        # This function was separated to be mocked in tests
        return cls.objects.create(**kwargs)

    @classmethod
    def _create_no_db(cls, **kwargs):
        idx = getattr(cls, '_no_db_index', 1) + 1
        return cls(id=idx, **kwargs)

    #
    # Instance methods
    #
    def __str__(self):
        class_name = type(self.object).__name__
        obj_data = f'{class_name}(id={self.object.id})'
        args = self.arguments.get('args', ())
        kwargs = self.arguments.get('kwargs', {})
        kwargs_data = (f'{key}={repr(value)}' for key, value in kwargs)
        args_data = ', '.join([obj_data, *args, *kwargs_data])
        return f'{self.function} ({self.status}): {args_data}'

    def wrap_result(self, result):
        """
        Convert result value into a raw valid database representation.

        The resulting value should be safe to store in the .result attribute.
        """
        return result

    def unwrap_result(self, raw):
        """
        Convert raw result value into a valid Python representation
        """
        return raw

    def set_args(self, obj, *args, **kwargs):
        """
        Sets arguments of the task function.
        """
        self.object = obj
        self.arguments = {'args': args, 'kwargs': kwargs}

    def call(self, *args, **kwargs):
        """
        Set object and extra arguments and run task in the foreground.
        """
        self.set_args(*args, **kwargs)
        return self.execute()

    def call_background(self, *args, **kwargs):
        """
        Set object and arguments and run task in the background.
        """
        self.set_args(*args, **kwargs)
        self.execute_background()
        return self

    def run(self, is_async=True):
        """
        Runs task, if necessary and fetch result.
        """
        if self.is_finished:
            return self if is_async else self.unwrap_result(self.result)
        elif self.is_pending:
            return self._run_pending(is_async)
        elif is_async:
            return self
        elif self.is_running:
            return self._wait_completion()
        elif self.is_failed:
            raise RuntimeError('task failed to complete')
        else:
            return None

    def _run_pending(self, is_async):
        if is_async:
            self.execute_background()
            return self
        else:
            result = self.execute()
            if self.is_failed:
                return self.run(is_async=False)
            return result

    def _wait_completion(self):
        dt = 0.01
        for __ in range(int(10 / dt)):
            sleep(dt)
            if not self.is_running:
                return self.run(is_async=False)
        raise TimeoutError('timeout while waiting for task completion')

    def execute(self, commit=True):
        """
        Run task synchronously and return the result.
        """

        self.status = self.STATUS.running
        if commit:
            self.save(update_fields=['status'])

        # Compute result
        result = None
        try:
            result = self.execute_function()
            log.info(f'task {self} successfully executed')
        except self.Skip:
            self.status = self.STATUS.skipped
            log.info(f'task skipped: {self}')
        except Exception as exc:
            self.status = self.STATUS.failed
            log.warning(f'error running task ({self}): {exc}')
        else:
            self.result = self.wrap_result(result)
            self.status = self.STATUS.finished
        if commit:
            self.save()
        return result

    def execute_background(self):
        """
        Schedule task to run in the background using celery.
        """
        task = self._get_handler(self.function).task

        if USE_CELERY:
            return task.delay(self.id)
        else:
            return task.apply([self.id])

    def execute_function(self):
        """
        Execute handler function and return its result.

        This function does not update the Task object.
        """
        args = self.arguments.get('args', ())
        kwargs = self.arguments.get('kwargs', {})
        function = self._get_handler(self.function).runner

        if self.get_object_model() == ContentType:
            object_id = kwargs.pop('object_id')
            model = self.object.model_class()
            object = model.objects.get(id=object_id)
        else:
            object = self.object

        return function(object, *args, **kwargs)

    def _get_handler(self, name):
        """
        Register a handler for some pair of (TaskClass, HandlerName).

        Handler names of a given task class are unique.
        """
        info = REGISTERED_FUNCTIONS[type(self), name]
        function = info['function']
        is_method = info['is_method']
        if is_method:
            return functools.partial(function, self)
        return function


class JsonTask(Task):
    """
    Special task that expect JSON results.
    """

    def wrap_result(self, result):
        import json

        return json.dumps(result)

    def unwrap_result(self, raw):
        import json

        return json.loads(raw)


class AbstractPandasTask(Task):
    """
    Abstract model for tasks that result in Pandas DataFrames.
    """
    result = models.BinaryField(blank=True)

    def wrap_result(self, result):
        try:
            return result.to_msgpack()
        except AttributeError:
            cls_name = result.__class__.__name__
            raise TypeError(f'result must be a DataFrame, got {cls_name}')

    def unwrap_result(self, raw):
        import pandas as pd

        return pd.read_msgpack(raw)

    class Meta:
        abstract = True



from .installed_apps import InstalledAppsConf


class MiddlewareConf(InstalledAppsConf):
    """
    Configure middleware classes.
    """

    with_middleware = InstalledAppsConf.with_app

    def get_middleware(self):
        apps = self.INSTALLED_APPS

        def add(middleware):
            if middleware.partition('.middleware')[0] in apps:
                return [middleware]
            else:
                return []

        return [
            'django.middleware.security.SecurityMiddleware',
            *add('django.contrib.sessions.middleware.SessionMiddleware'),
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            *add('django.contrib.auth.middleware.AuthenticationMiddleware'),
            *add('django.contrib.messages.middleware.MessageMiddleware'),
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
        ]



import enum


class _EnumDict(enum._EnumDict):
    """
    Namespace that automatically create value() instances for member entries.
    """

    def __init__(self, dtype=object):
        super().__init__()
        self.dtype = dtype
        self._descriptions = []

    def __setitem__(self, key, value):
        if key == 'description':
            raise ValueError('invalid enum name')
        super().__setitem__(key, value)

        if self._member_names and self._member_names[-1] == key:
            if not isinstance(value, tuple):
                if isinstance(value, self.dtype):
                    value = (value, key)
                elif isinstance(value, str):
                    value = (None, value)
                else:
                    value = (value, key)

            value, description = value
            self._last_values[-1] = value
            self._descriptions.append(description)
            dict.__setitem__(self, key, value)


class EnumMeta(enum.EnumMeta):
    dtype = object

    @classmethod
    def __prepare__(meta, cls, bases):  # noqa: N804
        enum_dict = _EnumDict(meta.dtype)
        member_type, first_enum = meta._get_mixins_(bases)
        if first_enum is not None:
            enum_dict['_generate_next_value_'] = getattr(first_enum, '_generate_next_value_', None)
        return enum_dict

    def __new__(meta, name, bases, namespace):  # noqa: N804
        values = namespace._last_values
        names = namespace._member_names

        # Update
        if meta.dtype is int:
            for idx, (key, value) in enumerate(zip(names, values)):
                values[idx] = value = idx if value is None else value
                dict.__setitem__(namespace, key, value)
            namespace._last_values = values

        new = super().__new__(meta, name, bases, namespace)
        return new

    def __init__(cls, name, bases, namespace):  # noqa: N805
        super().__init__(name, bases, namespace)

        # Save descriptions
        names = namespace._member_names
        descriptions = namespace._descriptions
        values = namespace._last_values

        cls._descriptions = dict(zip(names, descriptions))
        cls._values = dict(zip(names, values))
        for attr, descr in zip(names, descriptions):
            setattr(cls, attr + '_DESCRIPTION', descr)
            case = getattr(cls, attr)
            case.description = str(descr)
            cls._descriptions[case] = str(descr)

    def __hash__(cls):  # noqa: N805
        return id(cls)

    def get_description(cls, value):  # noqa: N805
        """
        Get description from value.
        """
        try:
            return cls._descriptions[value]
        except KeyError:
            raise ValueError('not a member of enumeration: %r' % value)


class IntEnum(enum.IntEnum,
              metaclass=type('IntEnumMeta', (EnumMeta,), {'dtype': int})):
    """
    A subclass of enum.IntEnum that accepts an optional human-friendly
    description field during declaration.

    It is safe to translate description strings.

    Usage:
        >>> class Roles(IntEnum):
        ...     TEACHER = 0, 'teacher'
        ...     STUDENT = 1, 'student'
    """


class Enum(enum.Enum, metaclass=EnumMeta):
    """
    Similar to :class:`boogie.IntEnum`, but accepts any type of value.
    """

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return other is self
        else:
            return self.value.__eq__(other)



from collections import OrderedDict, Sequence

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseBadRequest

from sidekick import lazy
from .route import Route, normalize_name, ModelLookupMixin, to_default_dict


# Django makes an instance check to see if urlpatterns is a list of path
# declarations. We inherit from list, but implement all methods using
# collections.Sequence. This is necessary since urls are only created after
# initialization of the Router.urls attribute.
class Router(ModelLookupMixin, Sequence, list):
    """
    A collection of routes.

    It exports a list of urlpatterns that can be included in a Django's url.py.
    """

    @lazy
    def urls(self):
        patterns = []
        pattern_groups = group_by_url(self.routes)
        for pattern, routes in pattern_groups.items():
            if len(routes) == 1:
                route, = routes
                patterns.append(route.path_handler())
            else:
                raise NotImplementedError(pattern)
        super().__setitem__(slice(None, None), patterns)
        return patterns

    def __init__(self, base_name='', template=None,
                 models=None, lookup_field=None, lookup_type=None, **kwargs):
        list.__init__(self)
        ModelLookupMixin.__init__(self, models, lookup_field, lookup_type)
        self.routes = []
        self.base_name = base_name
        self.template = template
        self.extra_args = kwargs

    def __call__(self, *args, **kwargs):
        return self.route(*args, **kwargs)

    def __len__(self):
        return len(self.urls)

    def __iter__(self):
        return iter(self.urls)

    def __getitem__(self, item):
        return self.urls[item]

    def route(self, path='', name=None, **kwargs):
        """
        Register a route from function. Users should provide a path and can
        provide any optional routing parameters.
        """

        def decorator(func):
            route = self.register(func, path, name, **kwargs)
            func.as_view = route.view_function
            return func

        return decorator

    def register(self, function, path='', name=None, template=None, **kwargs):
        """
        Register a function as a route.

        Similar to the .route method, but does not behave as a decorator.
        """

        # Use the last registered route if path is ellipsis
        if path is ... and not self.routes:
            msg = 'cannot determine last url from empty router'
            raise ImproperlyConfigured(msg)
        elif path is ...:
            path = self.routes[-1].path

        # Check if should use the parent template or not
        name = normalize_name(name, function)
        if template is None and self.template is not None:
            if isinstance(self.template, str):
                template = self.template.format(name=name)
            else:
                template = [x.format(name=name) for x in self.template]
        kwargs['name'] = name
        kwargs['template'] = template

        # Check if it override any model and lookup fields and types
        models = dict(kwargs.get('models') or ())
        kwargs['models'] = dict(self.models, **models)
        update_lookup(self, kwargs, 'lookup_field')
        update_lookup(self, kwargs, 'lookup_type')

        # Create a Route object
        kwargs = dict(self.extra_args, **kwargs)
        route = Route(path, function, **kwargs)
        self.routes.append(route)

        # Save route to the list of registered routes
        try:
            routes = getattr(function, 'routes', [])
            function.registered_routes = [*routes, route]
        except AttributeError:
            pass

        return route

    #
    # HTTP methods
    #
    def connect(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to CONNECT.
        """
        return self.route(*args, method='CONNECT', **kwargs)

    def delete(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to DELETE.
        """
        return self.route(*args, method='DELETE', **kwargs)

    def get(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to GET.
        """
        return self.route(*args, method='GET', **kwargs)

    def head(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to HEAD.
        """
        return self.route(*args, method='HEAD', **kwargs)

    def options(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to OPTIONS.
        """
        return self.route(*args, method='OPTIONS', **kwargs)

    def patch(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to PATCH.
        """
        return self.route(*args, method='PATCH', **kwargs)

    def post(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to POST.
        """
        return self.route(*args, method='POST', **kwargs)

    def put(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to PUT.
        """
        return self.route(*args, method='PUT', **kwargs)

    def trace(self, *args, **kwargs):
        """
        Similar to :method:`route`, but sets the HTTP method to TRACE.
        """
        return self.route(*args, method='TRACE', **kwargs)


def group_by_url(routes):
    dic = OrderedDict()
    for route in routes:
        paths = dic.setdefault(route.path, [])
        paths.append(route)
    return dic


def update_lookup(router, kwargs, field):
    original = getattr(router, field)
    new = kwargs.get(field, {})
    if isinstance(new, str):
        lookup = to_default_dict(new)
        lookup.update(original)
    else:
        lookup = original.copy()
        lookup.update(new)
    kwargs[field] = lookup


def multi_method_view(method_map):
    """
    Receives a mapping from HTTP method to view function and return a function
    dispatches to the correct implementation depending on the user requested
    method.
    """

    def view(request, **kwargs):
        try:
            handler = method_map[request.method]
        except KeyError:
            msg = f'method not allowed: {request.method}'
            return HttpResponseBadRequest(msg)
        return handler(request, **kwargs)

    return view



from enum import Enum
from functools import singledispatch, wraps

from django.apps import apps
from django.db import models
from rest_framework.decorators import action as action_decorator
from rest_framework.response import Response
from rest_framework.utils.encoders import JSONEncoder

from ..router.route import as_request_function
from ..utils.text import dash_case, humanize_name


def as_model(model):
    """
    Return a model class from model or string.
    """
    if isinstance(model, str):
        return apps.get_model(model)
    return model


def join_url(head, *args):
    """
    Join url parts. It prevents duplicate backslashes when joining url
    elements.
    """
    if not args:
        return head
    else:
        tail = join_url(*args)
        return f"{head.rstrip('/')}/{tail.lstrip('/')}"


def validation_error(err, status_code=403):
    """
    Return a JSON message describing a validation error.
    """
    errors = err.messages
    msg = {'status_code': status_code, 'error': True}
    if len(errors) == 1:
        msg['message'] = errors[0]
    else:
        msg['messages'] = errors
    return msg


def natural_base_url(model):
    """
    Return the natural base url name for the given model:
    * Uses a plural form.
    * Convert CamelCase to dash-case.
    """
    name = dash_case(model.__name__ + 's')
    return humanize_name(name).replace(' ', '-')


#
# Register converters to special Boogie types
#
def patch_rest_framework_json_encoder():
    """
    Patch rest_framework JSON encoder to accept objects that define a
    __json_default__ method that coerce data to JSON.
    """
    original = JSONEncoder.default

    if getattr(original, 'patched', False):
        return

    def default(self, obj):
        try:
            return original(self, obj)
        except TypeError:
            return to_json_default(obj)

    JSONEncoder.default = default
    default.patched = True
    default.original = original


@singledispatch
def to_json_default(obj):
    """
    Single dispatch function that register converters of arbitrary Python
    objects to JSON-compatible values.
    """

    if hasattr(obj, '__json_default__'):
        return obj.__json_default__()
    else:
        typename = obj.__class__.__name__
        raise TypeError(f"Object of type '{typename}' is not JSON serializable")


@to_json_default.register(Enum)
def enum_to_json(obj):
    return obj.name


#
# Viewset and serializer builders
#
def viewset_actions(actions):
    """
    Maps action names to decorated actions methods.
    """

    action_fields = {}
    for name, action in actions.items():
        func = action['method']
        args = action['args']
        action_fields[name] = action_method(func, name=name, **args)
    return action_fields


def action_method(function, is_method=False, detail=True, name=None, **kwargs):
    """
    Creates a new method decorated with a @action decorator to be inserted
    in a DRF viewset.
    """
    from boogie.rest import rest_api

    function = as_request_function(function)

    def wrap_result(request, result):
        if isinstance(result, Response):
            return result
        elif isinstance(result, (models.Model, models.QuerySet)):
            result = rest_api.serialize(result, request=request)
        return Response(result)

    def method(self, request, **kwargs):
        if is_method:
            result = function(self, request, **kwargs)
        elif detail:
            obj = self.get_object()
            result = function(request, obj)
        else:
            result = function(request)
        return wrap_result(request, result)

    method.__name__ = name or function.__name__
    if name:
        kwargs['url_path'] = name
    method = action_decorator(detail=detail, **kwargs)(method)
    return method


def with_model_cache(func):
    attr = '_%s_cache' % func.__name__

    @wraps(func)
    def method(self, model):
        if self.version is None:
            raise ValueError('cannot construct value if version is None')

        try:
            cache = getattr(self, attr)
        except AttributeError:
            cache = {}
            setattr(self, attr, cache)

        try:
            return cache[model]
        except KeyError:
            cache[model] = result = func(self, model)
            return result

    return method



from ..django_conf import Conf


class CeleryConf(Conf):
    CELERY_BROKER_URL = 'redis://localhost:6379/0'
    CELERY_ACCEPT_CONTENT = ['json']
    CELERY_TASK_SERIALIZER = 'json'
    CELERY_RESULT_SERIALIZER = 'json'



from rules import *
from rules.rulesets import default_rules as DEFAULT_RULES

from .proxy import proxy, proxy_seq
from .value import value
from .valuemap import ValueMap, compute, add_value, value_exists, remove_value, register_value, get_value

_NOT_GIVEN = object()


def register_rule(name, **kwargs):
    def decorator(func):
        pred = predicate(**kwargs)(func)
        add_rule(name, pred)
        return pred

    return decorator


def register_perm(name, **kwargs):
    def decorator(func):
        pred = predicate(**kwargs)(func)
        add_perm(name, pred)
        return pred

    return decorator


def get_rule(name, default=_NOT_GIVEN):
    try:
        return DEFAULT_RULES[name]
    except KeyError:
        if default is _NOT_GIVEN:
            raise ValueError('rule does not exist: %r' % name)
        return default



import logging
from operator import attrgetter
from warnings import warn

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.urls import path, include
from rest_framework import routers
from rest_framework.viewsets import ModelViewSet
from sidekick import lazy

from .api_info import ApiInfo
from .resource_info import ResourceInfo
from .utils import as_model, natural_base_url

log = logging.getLogger('boogie.rest_api')


class RestAPI:
    """
    Base class that stores global information for building an REST API with DRF.
    """
    router_class = routers.DefaultRouter

    @lazy
    def last_version(self):
        versions = set(self.api_registry)
        versions.remove(None)
        return max(versions) if versions else 'v1'

    @lazy
    def urls(self):
        versions = [v for v in self.api_registry if v is not None]
        patterns = self.get_urlpatterns
        return [
            *(path(f'{v}/', include(patterns(v))) for v in versions),
            path('', api_root_view(versions)),
        ]

    def __init__(self):
        self._api_info_base = api_info_base = ApiInfo(self)
        self.api_registry = {
            None: self._api_info_base,
            'v1': ApiInfo(self, version='v1', parent=api_info_base),
        }
        self.inlines_registry = {}

    def __call__(self, *args, version=None, inline=False, lookup_field='pk',
                 base_url=None, base_name=None, **kwargs):

        def decorator(cls):
            if not isinstance(cls, type) and not issubclass(type, models.Model):
                msg = f'must decorate a Django model subclass, got {cls}'
                raise TypeError(msg)
            kwargs.update(
                version=version,
                inline=inline,
                lookup_field=lookup_field,
                base_name=base_name,
                base_url=base_url,
            )
            self.register(cls, *args, **kwargs)
            return cls

        if len(args) == 1 and isinstance(args[0], type) and not kwargs:
            return decorator(args[0])
        return decorator

    def register(self, model, fields=None, *, version=None, inline=False, **kwargs):
        """
        Register class with the given meta data.

        Args:
            model:
                A Django model
            version:
                Optional API version string (e.g., 'v1'). If not given, it will
                register a resource to all API versions.
            fields:
                The list of fields used in the API. If not given, uses all
                fields.
            exclude:
                A list of fields that should be excluded.
            base_url:
                The base url address in which the resource is mounted. Defaults
                to a dashed case plural form of the model name.
            base_name:
                Base name for the router urls. Router will append suffixes such
                as <base_name>-detail or <base_name>-list. Defaults
                to a dashed case plural form of the model name.
            inline:
                Inline models are not directly part of an API, but can be
                embedded into other resources.

        Returns:
            An ResourceInfo object.
        """
        model = as_model(model)
        kwargs.update(inline=inline)
        resource_info = ResourceInfo(model, fields, **kwargs)
        info = self.get_api_info(version, create=True)
        info.register_resource(model, resource_info, inline=inline)
        return resource_info

    def register_viewset(self, viewset=None, base_url=None, *,  # noqa: C901
                         version='v1', model=None, skip_serializer=False):
        """
        Register a viewset class responsible for handling the given url.

        If a ModelViewSet is given, the viewset is automatically associated
        with a model and registered. Can be used as a decorator if the viewset
        argument is omitted.

        Args:
            viewset:
                Viewset subclass.
            base_url:
                Base url under which the viewset will be mounted. RestAPI can
                infer this URL from the model, when possible.
            version:
                API version name.
            model:
                Model associated with the viewset, when applicable.
            skip_serializer:
                If True, do not register serializer of ModelViewSet subclasses.
        """

        if isinstance(viewset, str) and base_url is None:
            base_url, viewset = viewset, None
        if viewset is None:
            args = locals()
            args.pop('self')
            args.pop('viewset')
            return lambda x: self.register_viewset(x, **args) or x

        api_info = self.get_api_info(version, create=True)

        # Discover the model class
        if model is None and issubclass(viewset, ModelViewSet):
            try:
                model = viewset.queryset.model
            except AttributeError:
                raise ImproperlyConfigured(
                    'could not determine the model of a ModelViewSet subclass. '
                    'Please pass the model explicitly when registering this '
                    'viewset.'
                )
        model = as_model(model)

        # Discover url
        if base_url is None and model:
            base_url = natural_base_url(model)
        if base_url is None:
            raise ImproperlyConfigured(
                'could not determine the base_url for this viewset. Please '
                'pass this parameter explicitly when registering the viewset.'
            )

        # Create ResourceInfo, if applicable
        if model is not None and skip_serializer:
            kwargs = {'base_url': base_url}

            def update(to, src):
                try:
                    kwargs[to] = src(viewset)
                except AttributeError:
                    pass

            update('fields', attrgetter('Meta.fields'))
            update('base_name', attrgetter('Meta.base_name'))
            update('lookup_field', attrgetter('lookup_field'))

            api_info[model] = ResourceInfo(model, **kwargs)

        # Register resource
        api_info.register_viewset(base_url, viewset)
        if not skip_serializer and issubclass(viewset, ModelViewSet):
            api_info.register_serializer(model, viewset.serializer_class)

    #
    # Decorators
    #
    def action(self, model, func=None, *, version=None, name=None, **kwargs):
        """
        Base implementation of both detail_action and list_action.

        Please use one of those specific methods.
        """

        def decorator(func):
            action_name = name or func.__name__.replace('_', '-')
            info.add_action(action_name, func, **kwargs)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    def detail_action(self, model, func=None, **kwargs):
        """
        Register function as an action for a detail view of a resource.

        Decorator that register a function as an action to the provided
        model.

        Args:
            model:
                A Django model or a string with <app_label>.<model_name>.
            func:
                The function that implements the action. It is a
                function that receives a model instance and return a response.
                RestAPI understands the following objects:

                * Django and DRF Response objects
                * A JSON data structure
                * An instance or queryset of a model that can be serialized by
                  the current API (it will serialize to JSON and return this
                  value)

                Exceptions are also converted to meaningful responses of the
                form ``{"error": true, "message": <msg>, "error_code": <code>}``.
                It understands the following exception classes:

                * PermissionError: error_code = 403
                * ObjectNotFound: error_code = 404
                * ValidationError: error_code = 400

                The handler function can optionally receive a "request" as
                first argument. RestAPI inspects function argument names to
                discover which form to call. This strategy may fail if your
                function uses decorators or other signature changing modifiers.
            version:
                Optional API version name.
            name:
                The action name. It is normally derived from the action function
                by simply replacing underscores by dashes in the function
                name.

        Usage:

            .. code-block:: python

                @rest_api.detail_action('auth.User')
                def books(user):
                    return user.user.books.all()

        This creates a new endpoint /users/<id>/books/ that displays all books
        for the given user.
        """
        return self.action(model, func, detail=True, **kwargs)

    def list_action(self, model, func=None, **kwargs):
        """
        Similar to :method:`detail_action`, but creates an endpoint associated
        with a list of objects.

        Usage:

            .. code-block:: python

                @rest_api.detail_action('auth.User')
                def books():
                    return Book.objects.filter(author__in=users)

            The new endpoint is created under /users/books/

        See Also:
            :meth:`detail_action`
        """
        return self.action(model, func, detail=False, **kwargs)

    def property(self, model, func=None, *, version='v1', name=None):
        """
        Decorator that declares a read-only API property.

        Args:
            model:
                The model name.
            version:
                API version. If omitted, it will be included in all API
                versions.
        """

        def decorator(func):
            prop_name = name or func.__name__
            info.add_property(prop_name, func)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    def link(self, model, func=None, *, version='v1', name=None):
        """
        Decorator that declares a function to compute a link included into the
        "links" section of the serialized model.

        Args:
            model:
                The model name.
            version:
                API version. If omitted, it will be included in all API
                versions.
        """

        def decorator(func):
            link_name = name or func.__name__
            info.add_link(link_name, func)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    #
    # Hooks
    #
    def save_hook(self, model, func=None, *, version='v1'):
        """
        Decorator that registers a hook that is executed when a new object is
        about to be saved. This occurs both during object creation and when it
        is updated. The provided function receives a request and an unsaved
        instance as arguments and must save the instance to the database and
        return it.

        Args:
            model:
                The model name.
            version:
                API version. If omitted, it will be included in all API
                versions.

        Examples:

            .. code-block:: python

                @rest_api.save_hook(Book)
                def save_book(request, book):
                    book.save()  # Don't forget saving the instance!
                    book.owner = request.user
                    return book
        """

        def decorator(func):
            info.add_hook('save', func)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    def delete_hook(self, model, func=None, *, version='v1'):
        """
        Decorator that registers a hook that is executed before a new object is
        about to be deleted.

        Deletion can be prevented either by raising an exception (which will
        generate an error response) or silently by not calling the .delete()
        method of a model or queryset.

        Args:
            model:
                The model name.
            version:
                API version. If omitted, it will be included in all API
                versions.

        Examples:

            .. code-block:: python

                @rest_api.delete_hook(Book)
                def delete_book(request, book):
                    if book.user_can_remove(request.user):
                        book.delete()
                    else:
                        raise PermissionError('user cannot delete book!')
        """

        def decorator(func):
            info.add_hook('delete', func)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    def query_hook(self, model, func=None, *, version='v1'):
        """
        Decorator that registers a hook that is executed to extract the
        queryset used by the viewset class.

        Args:
            model:
                The model name.
            version:
                API version. If omitted, it will be included in all API
                versions.

        Examples:

            .. code-block:: python

                @rest_api.query_hook(Book)
                def query_hook(request, qs):
                    return qs.all()
            """

        def decorator(func):
            info.add_hook('query', func)
            return func

        info = self.get_resource_info(model, version)
        return decorator if func is None else decorator(func)

    #
    # Actions
    #
    def serialize(self, obj, request=None, version='v1'):
        """
        Serialize object and return the corresponding JSON structure.
        """

        if isinstance(obj, models.Model):
            model = type(obj)
            many = False
        else:
            model = obj.model
            many = True

        ctx = {'request': request} if request is not None else None
        serializer = self.get_serializer(model, version=version)
        result = serializer(obj, many=many, context=ctx)
        return result.data

    def get_hyperlink(self, obj, request=None, version='v1'):
        """
        Return the hyperlink of the given object in the API.
        """
        info = self.get_resource_info(type(obj), version=version)
        return info.detail_hyperlink(obj, request, version)

    def get_router(self, version='v1'):
        """
        Gets a DRF router object for the given API version.

        Args:
            version: An API version string.
        """
        api_info = self.get_api_info(version)
        router = self.router_class()
        router.root_view_name += '-' + version
        entries = sorted(api_info.iter_viewset_items())

        # Registered sorted entries
        for url, viewset in entries:
            base_name = getattr(viewset, 'base_name', None)
            router.register(url, viewset, base_name)
            log.debug('created viewset %s at %s' % (url, base_name))
        return router

    def get_urls(self, version='v1'):
        warn('this function is deprecated, please use get_urlpatterns instead.')
        return self.get_router(version).urls

    def get_urlpatterns(self, version='v1'):
        """
        Return a list of urls to be included in Django's urlpatterns::

        Usage:
            .. code-block:: python

                urlpatterns = [
                    ...,
                    path('api/v1/', include(rest_api.get_urlpatterns('v1')))
                ]

        See Also:
            :meth:`get_router`
        """
        return self.get_router(version).urls

    def get_serializer(self, model, version='v1'):
        """
        Return the serializer class for the given model.
        """
        api_info = self.get_api_info(version=version)
        return api_info.serializer_class(model)

    def get_viewset(self, model, version='v1'):
        """
        Return the viewset class for the given model.
        """
        api_info = self.get_api_info(version=version)
        return api_info.viewset_class(model)

    def get_api_info(self, version='v1', create=False):
        """
        Return the ApiInfo instance associated with the given API version.

        If version does not exist and create=True, it creates a new empty
        ApiInfo object.

        Returns an :class:`ApiInfo` instance.
        """
        try:
            registry = self.api_registry[version]
        except AttributeError:
            if not create:
                raise
            info_base = self._api_info_base
            registry = ApiInfo(self, version=version, parent=info_base)
            self.api_registry[version] = registry
        return registry

    def get_resource_info(self, model, version='v1'):
        """
        Return the resource info object associated with the given model. If
        version does not exist, create a new ApiInfo object for the given
        version.

        Args:
            model:
                A model class or a string in the form of 'app_label.model_name'
            version:
                Version string or None for the default api constructor.

        Returns:
            A :class:`ResourceInfo` instance.
        """
        model = as_model(model)
        registry = self.get_api_info(version, create=True)
        try:
            return registry[model]
        except KeyError:
            model_name = model._meta.name
            version_string = '' if not version else ' (%s)' % version
            raise ImproperlyConfigured(
                '{model} is not registered on the API{version}. Please '
                'decorate the model class with the boogie.rest.rest_api() '
                'decorator.'.format(model=model_name, version=version_string)
            )


#
# Utility functions and classes
#
def api_root_view(versions, description=None):
    class RestApiView(routers.APIRootView):
        """
        This is the base entry point for REST resources and provides access to
        all versions of the REST API.
        """

    if description:
        RestApiView.__doc__ = description

    views = {version: 'api-root-' + version for version in versions}
    return RestApiView.as_view(api_root_dict=views)



import base64
import importlib.util
from hashlib import md5


def secret_hash(data):
    """
    Create a secret hash from data.
    """
    strings = []
    for key, value in sorted(data.items()):
        strings.append(key)
        try:
            if isinstance(value, dict):
                value = sorted(value.items())
            if isinstance(value, list):
                value = tuple(value)
            data = hash(value)
            if data != -1:
                strings.append(str(data))
        except TypeError:
            pass
    data = ''.join(strings)
    hash_value = md5(data.encode('utf8')).digest()
    return base64.b85encode(hash_value).decode('ascii')


def module_exists(mod, package=None):
    spec = importlib.util.find_spec(mod, package=package)
    return spec is not None


def module_path(mod, package=None):
    spec = importlib.util.find_spec(mod, package=package)
    return spec.origin



import rules as _rules


class Rules:
    """
    Implement the "rules" attribute of a model.

    This attribute decouples all business rules regarding authorization from the
    model class. The goal is to implement thinner models.
    """

    def __init__(self, model=None):
        self.model = model

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        if self.model:
            return self
        return Rules(instance)

    def _test_multilple(self, user, rule_list, reducer, test):
        """
        Common implementation to test() and test_any()
        """
        model = self.model
        return user.is_superuser or \
            reducer(test(rule, user, model) for rule in rule_list)

    def test(self, user, *rules):
        """
        Return True if user has all given permissions.
        """
        return self._test_multilple(user, rules, all, _rules.test_rule)

    def test_any(self, user, *rules):
        """
        Return True if user has any of the given permissions.
        """
        return self._test_multilple(user, rules, any, _rules.test_rule)

    def has_perm(self, user, *rules):
        """
        Return True if user has all given permissions.
        """
        return self._test_multilple(user, rules, all, _rules.has_perm)

    def has_perm_any(self, user, *rules):
        """
        Return True if user has any of the given permissions.
        """
        return self._test_multilple(user, rules, any, _rules.has_perm)



import html
import io
from functools import singledispatch

from django.conf import settings
from django.http import HttpResponseServerError, HttpResponse


@singledispatch
def render_response(obj):
    """
    Convert arbitrary Python objects to http responses.

    By default, it tries to render object using :func:`render_html`. If it
    cannot be rendered that way and no handler was specified for the object
    type, it returns a :class:`django.http.HttpResponseServerError` response.
    """
    try:
        html = render_html(obj)
    except TypeError:
        if settings.DEBUG:
            msg = f'bad result from a view function: {repr(obj)}'
            return HttpResponseServerError(msg)
        else:
            return HttpResponseServerError()
    else:
        return HttpResponse(html)


@render_response.register(str)
def _(st):
    return HttpResponse(html.escape(st))


def render_html(obj):
    """
    Renders object as an HTML string in the given file.

    The function should write to the
    """
    handler = dump_html.dispatch(type(obj))
    file = io.StringIO()
    handler(obj, file)
    return file.getvalue()


@singledispatch
def dump_html(obj, file):
    """
    Write HTML representation of object in the given file handler.
    """
    type_name = obj.__class__.__name__
    raise TypeError(f'cannot convert {type_name} to html')


#
# Register hyperpython, if available
#
try:
    from hyperpython import Element, Text
except ImportError:
    pass
else:
    @dump_html.register(Element)
    @dump_html.register(Text)
    def _(x, fd):
        x.dump(fd)

    @render_response.register(Element)
    @render_response.register(Text)
    def _(x):
        return HttpResponse(str(x))



def humanize_name(name):
    """
    "Humanize" camel case or Python variable name.

    Usage:
        >>> humanize_name('SomeName')
        'some name'
    """
    name = name.replace('_', ' ')
    parts = []
    buffer = []
    is_last_lower = False
    for chr in name:
        if chr.isupper():
            if is_last_lower:
                parts.append(''.join(buffer))
                buffer = [chr]
            else:
                buffer.append(chr)
            is_last_lower = False
        else:
            is_last_lower = True
            buffer.append(chr)

    parts.append(''.join(buffer))
    return ' '.join(parts)


def plural(st):
    """
    Convert string into a probable plural form.
    """
    return st + 's'


def indent(text, indent='    '):
    """
    Indent text by the given indentation string.
    """
    return '\n'.join(indent + line for line in text.splitlines())


def safe_repr(obj, max_length=None, repr=repr):
    """
    A safe version of repr that is guaranteed to never raise exceptions.
    """
    # noinspection PyBroadException
    try:
        data = repr(obj)
    except Exception as ex:
        cls_name = type(obj).__name__
        ex_name = type(ex).__name__
        data = f'<{cls_name} object [{ex_name}: {ex}]>'

    if max_length is not None and len(data) > max_length:
        data = data[:max_length - 3] + '...'

    return data


def first_line(st):
    """
    Extracts first line of string.
    """
    return st.lstrip().partition('\n')[0]


def dash_case(name):
    """
    Convert a camel case string to dash case.

    Example:
        >>> dash_case('SomeName')
        'some-name'
    """

    letters = []
    for c in name:
        if c.isupper() and letters and letters[-1] != '-':
            letters.append('-' + c.lower())
        else:
            letters.append(c.lower())
    return ''.join(letters).replace('_', '-')


def snake_case(name):
    """
    Convert camel case to snake case.
    """
    return dash_case(name).replace('-', '_')



import os

from invoke import task


@task
def clean(ctx):
    """
    Clean pyc files and build assets.
    """
    join = os.path.join
    rm_files = []
    rm_dirs = []
    for base, subdirs, files in os.walk('.'):
        if '__pycache__' in subdirs:
            rm_dirs.append(join(base, '__pycache__'))
        elif os.path.basename(base) == '__pycache__':
            rm_files.extend(join(base, f) for f in files)

    print('Removing compiled bytecode files')
    for path in rm_files:
        os.unlink(path)
    for path in rm_dirs:
        os.rmdir(path)



import sidekick as sk

from django import http

BOOGIE_VIEW_MIDDLEWARES = {}
NOTGIVEN = object()


def groups_required(groups, **kwargs):
    """
    Return a middleware that validates if the current user belongs to all of
    the provided groups.
    """
    kwargs.setdefault('message', 'Insufficient permissions')
    return access_control(lambda r: r.user.has_groups(groups), **kwargs)


def perms_required(perms, *, fn=None, **kwargs):
    """
    Return a middleware that validates if the current user has the given
    permissions.
    """
    kwargs.setdefault('message', 'Insufficient permissions')
    if fn:
        return access_control(lambda r: r.user.has_perms(perms, fn(r)), **kwargs)
    else:
        return access_control(lambda r: r.user.has_perms(perms), **kwargs)


def staff_required(**kwargs):
    """
    Return a middleware that validates if the current user is staff.
    """
    kwargs.setdefault('message', 'Insufficient permissions')
    return access_control(lambda r: r.user.is_staff, **kwargs)


def login_required(login_url=None, **kwargs):
    """
    Return a middleware that validates if the current user is authenticated.
    """
    if login_url is False:
        kwargs.setdefault('message', 'Requires login')
        return access_control(lambda r: r.user.is_authenticated, **kwargs)

    from django.contrib.auth.decorators import login_required

    def middleware(get_response):
        return login_required(login_url=login_url, **kwargs)(get_response)

    return middleware


def check_rule(rule, fn: callable = None, **kwargs):
    """
    Return a middleware that validates if the current user respects the
    given rule.
    """
    import rules

    if fn is None:
        test = (lambda r: rules.test_rule(rule, r.user))
    else:
        test = (lambda r: rules.test_rule(rule, r.user, fn(r)))
    return access_control(test, **kwargs)


def access_control(grant_access,
                   response_type=http.HttpResponseForbidden,
                   message='Not allowed'):
    def middleware(get_response):
        def handler(request):
            if grant_access(request):
                return get_response(request)
            return response_type(message)

        return handler

    return middleware


def unpoly_middleware(get_response):
    """
    Add unpoly headers to request and response objects.
    """

    def handler(request):
        # Read request headers and creates the request.up object
        value = request.META.get
        request.up = sk.record(
            location=value('X-Up-Location'),
            target=value('X-Up-Target'),
            fail_target=value('X-Up-Fail-Target'),
            validate=value('X-Up-Validate'),
        )

        # Handle response and extract title from the context
        response = get_response(request)
        if hasattr(response, 'context'):
            title = request.context.get('title')
            if title:
                response.META.setdefault('X-Up-Title', title)

        return response

    return handler


BOOGIE_VIEW_MIDDLEWARES.update({
    'staff_required': staff_required(),
    'login_required': login_required(),
    'unpoly': unpoly_middleware,
})



from celery import shared_task
from django.core.exceptions import ImproperlyConfigured

from .settings import REGISTERED_FUNCTIONS


def task_method(func):
    """
    Register method as a task.
    """
    func.is_task = True
    func.is_method = True
    return func


def task(cls, name=None, function=None):
    """
    Register a handler function that executes a task.

    Each task should be associated with a Job class.
    """
    if not isinstance(cls, type):
        raise ImproperlyConfigured(
            'Job class is missing as decorator argument. Please fix your '
            'decorator to be like @task(JobClass)'
        )
    if function is None:
        return lambda function: task(cls, name, function) or function

    name = name or function.__name__
    task_name = f'{cls.__module__}.{cls.__name__}:{name}'

    def call(object, *args, **kwargs):
        return cls.objects.call(name, object, *args, **kwargs)

    def delayed(object, *args, **kwargs):
        return cls.objects.delayed(name, object, *args, **kwargs)

    def schedule(object, *args, **kwargs):
        return cls.objects.schedule(name, object, *args, **kwargs)

    @shared_task(name=task_name)
    def celery_task(task_id):
        task = cls.objects.get(id=task_id)
        return task.run(is_async=False)

    function.task_name = task_name
    function.call = call
    function.delayed = delayed
    function.schedule = schedule
    function.celery_runner = celery_task

    REGISTERED_FUNCTIONS[cls, name] = {
        'function': function,
        'is_method': getattr(function, 'is_method', False),
    }



from collections import defaultdict, MutableMapping

CONF_PARAMS = defaultdict(dict)
GLOBAL_PARAMS = defaultdict(dict)


class Params(MutableMapping):
    """
    A dictionary of parameters.

    Keys are searched first in 4 stores (in order):

    * local store
    * configuration store
    * global store
    * defaults values store
    """

    def __init__(self, id, local=(), defaults=()):
        self._id = id
        self._local = dict(local)
        self._conf = CONF_PARAMS[id]
        self._global = GLOBAL_PARAMS[id]
        self._defaults = dict(defaults)
        self._stores = [self._local, self._conf, self._global, self._defaults]

    def __repr__(self):
        return repr(dict(self))

    def __call__(self, **kwargs):
        new_local = dict(self._local, **kwargs)
        return Params(id, new_local, self._defaults)

    def __getitem__(self, key):
        for store in self._stores:
            if key in store:
                return store[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._local[key] = value

    def __delitem__(self, key):
        if key in self._local:
            del self._local[key]
        elif key in self._defaults:
            del self._defaults[key]
        raise KeyError(key)

    def __len__(self):
        return sum(1 for _ in self)

    def __iter__(self):
        used = set()
        for store in self._stores:
            for key in store:
                if key not in used:
                    yield key
                else:
                    used.add(key)


def get_params(id, **kwargs):
    """
    Returns a mapping of parameters defined either in settings.py or by a call
    to the :func:`set_params` function.
    """
    return Params(id, defaults=kwargs)


def set_params(id, **kwargs):
    """
    Declares a mapping of parameters and sets the global values for its keys.
    """
    GLOBAL_PARAMS[id].update(kwargs)
    return Params(id)


def update_configuration(dic):
    """
    Update the CONF_PARAMS variable inplace with the configuration dictionary
    that is usually defined in settings.py.

    Return the update CONF_PARAMS dictionary.
    """

    for id, params in dic.items():
        CONF_PARAMS[id].update(params)

    return CONF_PARAMS



import functools
import inspect
import re
from collections import defaultdict

from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse, HttpResponseForbidden, Http404
from django.shortcuts import render
from django.urls import path
from django.views.decorators.cache import never_cache, cache_control
from django.views.decorators.clickjacking import xframe_options_exempt, xframe_options_deny, xframe_options_sameorigin
from django.views.decorators.csrf import csrf_exempt, ensure_csrf_cookie, requires_csrf_token, csrf_protect
from django.views.decorators.gzip import gzip_page

from .paths import register_model_converter, get_lookup_type
from ..render import render_response


class Cte:
    def __init__(self, cte):
        self.value = cte

    def __call__(self, *args, **kwargs):
        return self.value

    def __repr__(self):
        return 'Cte(%r)' % self.value


class ModelLookupMixin:
    """
    Common initialization patterns for Route and Router.
    """

    def __init__(self, models, lookup_field, lookup_type):
        # Normalize models
        self.models = dict(models or ())
        if isinstance(lookup_field, defaultdict):
            self.lookup_field = lookup_field.copy()
        else:
            self.lookup_field = to_default_dict(lookup_field or {}, 'pk')
        if isinstance(lookup_type, defaultdict):
            self.lookup_type = lookup_type.copy()
        else:
            self.lookup_type = to_default_dict(lookup_type or {}, 'str')


class Route(ModelLookupMixin):
    """
    A route is a combination of a Django view + some url pattern.
    """

    def __init__(self, path, func, name=None, method=None, template=None,
                 login=False, perms=None, staff=False, object=None,
                 cache=None, gzip=False, xframe=None, csrf=None,
                 models=None, lookup_field='pk', lookup_type=None,
                 decorators=()):
        super().__init__(models, lookup_field, lookup_type)
        self.path = path
        self.function = func
        self.name = normalize_name(name, func)
        self.method = method
        self.template = template
        self.login = login
        self.perms = perms
        self.object = object
        self.staff = staff
        self.cache = cache
        self.gzip = gzip
        self.xframe = xframe
        self.csrf = csrf
        self.decorators = tuple(decorators)

    def view_function(self):
        """
        Return a function that respects Django's view function contract.
        """

        function = as_request_function(self.function)
        decorators = ['login', 'staff', 'perms',
                      'cache', 'gzip', 'xframe', 'csrf',
                      'decorators']
        if self.object:
            decorators.remove('perms')
        kwargs = {attr: getattr(self, attr) for attr in decorators}

        @apply_decorators(**kwargs)
        def view_function(request, **kwargs):
            try:
                self.prepare_arguments(request, kwargs)
                result = function(request, **kwargs)
                return self.prepare_response(result, request)
            except PermissionError as exc:
                return HttpResponseForbidden(str(exc))

        return view_function

    def path_handler(self, prefix=''):
        """
        Returns a django.urls.path (or re_path) object that handles the given
        route.
        """
        path_ = self.compatible_path()
        return path(path_, self.view_function(), name=prefix + self.name)

    def prepare_response(self, result, request):
        """
        Return a Django response object from the result of the supplied view
        function.
        """
        # Regular http responses
        if isinstance(result, HttpResponse):
            return result

        # Template context
        elif isinstance(result, dict) and self.template is not None:
            return render(request, self.template, result)

        # Other types
        return render_response(result)

    def prepare_arguments(self, request, args):
        """
        Transforms the dictionary of input arguments inplace.

        This step transforms the view arguments before passing them to the view
        function.
        """
        if self.object and self.perms:
            obj = args.get(self.object)
            if obj is not None:
                user = request.user
                for perm in self.perms:
                    if not user.has_perm(perm, obj):
                        raise Http404

        for argname in self.models:
            if argname in args and args[argname] is None:
                raise Http404('%s not found' % argname)

    def compatible_path(self):
        """
        Convert a Boogie-style path specification to a valid Django path.
        """
        path = self.path

        for name, model in self.models.items():
            part = f'<model:{name}>'
            if part in path:
                lookup_field = self.lookup_field[name]
                lookup_type = self.lookup_type[name]
                if lookup_type is None:
                    lookup_type = get_lookup_type(None, model, lookup_field)
                converter = get_converter(model, lookup_field, lookup_type)
                path = path.replace(part, f'<{converter}:{name}>')

        # Find unregistered models
        if '<model:' in path:
            m = re.search(r'<model:[^>]*>', path)
            part = path[m.start():m.end()]
            name = part[7:-1]
            raise ImproperlyConfigured(
                f'Could not find a model for {part}. Please pass the correct '
                f'model for {name} in the models dictionary of the route.'
            )
        return path


#
# Auxiliary functions
#
def as_request_function(function):
    """
    Inspect to check if function receives a request as first parameter and
    return either the original function (if it expects a request) or a
    transformed function that omits the first parameter.
    """
    spec = inspect.getfullargspec(function)

    if spec.args and spec.args[0] == 'request':
        return function

    @functools.wraps(function)
    def request_function(request, *args, **kwargs):
        return function(*args, **kwargs)

    return request_function


def apply_decorators(view=None, login=False, staff=False, perms=None,  # noqa: C901
                     cache=None, gzip=False, xframe=None, csrf=None,
                     decorators=()):
    """
    Apply decorators to view function. Can also be used as a decorator.
    """

    if view is None:
        kwargs = locals()
        kwargs.pop('view')
        return lambda view: apply_decorators(view, **kwargs)

    # Cache control
    if cache is False:
        view = never_cache(view)
    elif cache is not None:
        view = cache_control(**cache)(view)

    # Permissions
    if login:
        view = login_required(view)
    if perms:
        view = permission_required(perms)(view)
    if staff:
        view = staff_required(view)

    # Compression
    if gzip:
        view = gzip_page(view)

    # Security
    if xframe is False:
        view = xframe_options_exempt(view)
    elif xframe == 'deny':
        view = xframe_options_deny(view)
    elif xframe == 'sameorigin':
        view = xframe_options_sameorigin(view)
    if csrf is False:
        view = csrf_exempt(view)
    elif csrf == 'cookie':
        view = ensure_csrf_cookie(view)
    elif csrf == 'token':
        view = requires_csrf_token(view)
    elif csrf is True:
        view = csrf_protect(view)

    # Apply final decorators
    for decorator in reversed(decorators):
        view = decorator(view)
    return view


def staff_required(view):
    @functools.wraps(view)
    def decorated(request, **kwargs):
        if not request.user.is_staff:
            return HttpResponseForbidden()
        return view(request, **kwargs)

    return decorated


def normalize_lookup(field, name=None):
    if isinstance(field, str):
        return field
    return field.get(name)


def normalize_name(name, function=None):
    if name:
        return name
    return function.__name__.replace('_', '-')


def to_default_dict(value, default=None):
    if isinstance(value, defaultdict):
        return value.copy()
    elif isinstance(value, dict):
        if default is not None:
            data = defaultdict(Cte(default))
            data.update(value)
            return data
        else:
            return dict(value)
    else:
        return defaultdict(Cte(value))


@functools.lru_cache(maxsize=256)
def get_converter(model, lookup_field, lookup_type):
    name = f'${model.__name__}.{lookup_field}-{lookup_type}'
    register_model_converter(model, name, lookup_field, lookup_type)
    return name



from boogie.apps.tasks.settings import REGISTERED_FUNCTIONS


class TaskProperty:
    """
    Implements the Task.task.<method> interface.
    """

    def __init__(self, cls):
        self._cls = cls

    def __call__(self, func=None, **kwargs):
        if func is None:
            return lambda func: self(func, **kwargs)

        raise NotImplementedError

    def __getattr__(self, attr):
        funcs = get_functions(self._cls)
        try:
            return funcs[attr]
        except KeyError:
            raise AttributeError(f'function not registerd: {attr}')


class TaskPropertyDescriptor:
    def __get__(self, instance, owner=None):
        if instance is not None:
            raise ValueError(
                'task attribute cannot be accessed from task instances'
            )
        return TaskProperty(owner)


def get_functions(cls):
    items = REGISTERED_FUNCTIONS
    return {f for cls_, f in items if cls is cls_}


import os
import sys

from invoke import task

python = sys.executable


@task
def run(ctx, no_toolbar=False):
    """
    Run development server
    """
    env = {}
    if no_toolbar:
        env['DISABLE_DJANGO_DEBUG_TOOLBAR'] = 'true'
    else:
        manage(ctx, 'runserver 0.0.0.0:8000', env=env)


@task
def gunicorn(ctx, debug=None, environment='production', port=8000, workers=4):
    """
    Run application using gunicorn for production deploys.

    It assumes that static media is served by a reverse proxy.
    """

    from gunicorn.app.wsgiapp import run as run_gunicorn
    directory = ''

    env = {
        'DISABLE_DJANGO_DEBUG_TOOLBAR': str(not debug),
        'PYTHONPATH': 'src',
        'DJANGO_ENVIRONMENT': environment,
    }
    if debug is not None:
        env['DJANGO_DEBUG'] = str(debug).lower()
    os.environ.update(env)
    args = [
        'ej.wsgi', '-w', str(workers), '-b', f'0.0.0.0:{port}',
        '--error-logfile=-',
        '--access-logfile=-',
        '--log-level', 'info',
        f'--pythonpath={directory}/src'
    ]
    sys.argv = ['gunicorn', *args]
    run_gunicorn()


@task
def clean_migrations(ctx, all=False, yes=False):
    """
    Remove all automatically created migrations.
    """
    import re
    auto_migration = re.compile(r'\d{4}_auto_\w+.py')
    all_migration = re.compile(r'\d{4}\w+.py')

    rm_list = []
    for app in os.listdir('src'):
        migrations_path = f'src/{app}/migrations/'
        if os.path.exists(migrations_path):
            migrations = os.listdir(migrations_path)
            if '__pycache__' in migrations:
                migrations.remove('__pycache__')
            if all:
                rm_list.extend([
                    f'{migrations_path}{f}' for f in migrations
                    if all_migration.fullmatch(f)
                ])
            elif sorted(migrations) == ['__init__.py', '0001_initial.py']:
                rm_list.append(f'{migrations_path}/0001_initial.py')
            else:
                rm_list.extend([
                    f'{migrations_path}/{f}' for f in migrations
                    if auto_migration.fullmatch(f)
                ])
    remove_files(rm_list, yes)


def remove_files(remove_files, yes):
    print('Listing auto migrations')
    for file in remove_files:
        print(f'* {file}')
    if all:
        print('REMOVING ALL MIGRATIONS IS DANGEROUS AND SHOULD ONLY BE '
              'USED IN TESTING')
    if yes or input('Remove those files? (y/N)').lower() == 'y':
        for file in remove_files:
            os.remove(file)


#
# DB management
#
@task
def db(ctx, migrate_only=False):
    """
    Perform migrations
    """
    if not migrate_only:
        manage(ctx, 'makemigrations')
    manage(ctx, 'migrate')


@task
def db_reset(ctx):
    """
    Reset data in database and optionally fill with fake data
    """
    ctx.run('rm -f local/db/db.sqlite3')
    manage(ctx, 'migrate')


@task
def db_fake(ctx, users=True, conversations=True, admin=True, safe=False, theme=None):
    """
    Adds fake data to the database
    """
    set_theme(theme)
    msg_error = 'Release build. No fake data will be created!'

    if safe:
        if os.environ.get("FAKE_DB") == 'true':
            print('Creating fake data...')
        else:
            return print(msg_error)
    if users:
        manage(ctx, 'createfakeusers', admin=admin)
    if conversations:
        manage(ctx, 'createfakeconversations')


#
# Utility functions
#
def manage(ctx, cmd, env=None, **kwargs):
    """
    Call python manage.py in a more robust way
    """
    kwargs = {k.replace('_', '-'): v for k, v in kwargs.items() if v is not False}
    opts = ' '.join(f'--{k} {"" if v is True else v}' for k, v in kwargs.items())
    cmd = f'{python} manage.py {cmd} {opts}'
    env = {**os.environ, **(env or {})}
    path = env.get("PYTHONPATH", ":".join(sys.path))
    env.setdefault('PYTHONPATH', f'src:{path}')
    ctx.run(cmd, pty=True, env=env)


set_theme = (lambda *args: None)



import functools
from pathlib import Path

import environ

path_type = type(Path('path'))


def env(default, type=None, name=None, **kwargs):
    """
    Declare a value that can be overridden by environment variables.

    Usage:
        Used in class declarations:

        class Conf:
            THE_ANSWER = env(42)

        Now Conf instances have an THE_ANSWER attribute that has a default value
        of 42, but can be overridden by a "THE_ANSWER" environment variable.
        If the configuration class define a ``env_prefix`` attribute, it will be
        prepended to the name of the environment variable.

    Args:
        default:
            The default value for the variable, if not present in the
            environment.
        type:
            Variable type. Usually this can be inferred from the default value.
            Boogie also accept some special values described bellow.
        name:
            Explicit name of the environment variable. By default, the name is
            derived from ``<class>.env_prefix + <attribute_name>``.

    Notes:
        Boogie understands the following builtin types: str, bool, int, float,
        list, dict and tuple. It also understands the following pseudo-types:

        json:
            Environment holds arbitrary JSON data.
        path:
            Path to a valid file in the filesystem. Returns a
            :class:`pathlib.Path` instance.
        db_url:
            URL to a database connection, e.g., sqlite:///db.sqlite3 or
            psql://user:password@domain:port/database
        cache_url:
            URL to cache provider, e.g., memcache://host:port
        search_url:
            URL to search provider, e.g., elasticsearch://host:port
        email_url:
            URL to e-mail provider, e.g., smtp+ssl://host:port

    Returns:
        A env variable descriptor
    """
    return EnvDescriptor(default, type=type, name=name, **kwargs)


def env_settings(**kwargs):
    """
    Configure information about environment variable associated with the given
    method. This controls how the "env" attribute is exposed to the function.

    Args:
        name:
            Name of the environment variable.
        type:
            Type (str, int, float, bool, etc).
    """
    valid_kwargs = {'name', 'type', 'default'}

    def decorator(func):
        for k, v in kwargs.items():
            if k not in valid_kwargs:
                raise TypeError(f'invalid argument: {k}')
            setattr(func, f'env_{k}', v)
        return func

    return decorator


def env_default(**options):
    """
    Use env variable, if set, otherwise execute function to compute attribute.
    """
    not_given = object()

    def decorator(func):
        name = getattr(func, 'env_name', func.__name__)
        type = getattr(func, 'env_type', str)
        default = getattr(func, 'env_default', not_given)

        @functools.wraps(func)
        def decorated(self):
            value = self.env(name, type=type, default=default)
            if value is not_given:
                return func(self)
            return value

        return decorated

    return decorator


class EnvDescriptor:
    METHOD_MAPPER = {
        # Basic Python types
        str: 'str',
        bool: 'bool',
        float: 'float',
        int: 'int',
        list: 'list',
        tuple: 'tuple',
        dict: 'dict',
        path_type: 'str',

        # Special methods
        'json': 'json',
        'path': 'path',
        'db_url': 'db_url',
        'cache_url': 'cache_url',
        'search_url': 'search_url',
        'email_url': 'email_url',
    }

    def __init__(self, default, type=None, name=None, **kwargs):
        self.default = default
        self.name = name
        self.kwargs = kwargs

        if type is None:
            type = default.__class__
        if type not in self.METHOD_MAPPER:
            raise ValueError(f'invalid type: {type}')
        self.type = type

    def __set_name__(self, owner, name):
        prefix = getattr(owner, 'env_prefix', '')
        if self.name is None:
            self.name = prefix + name
        elif '{' in self.name:
            self.name = name.format(prefix=prefix, attr=name)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        value = self.get_value(instance)
        setattr(instance, self.name, value)
        return value

    def get_value(self, conf):
        method = self.get_env_method(conf)
        return method(self.name, default=self.default, **self.kwargs)

    def get_env_method(self, conf):
        method = self.METHOD_MAPPER[self.type]
        return getattr(conf.env, method)


class EnvProperty(EnvDescriptor):
    def __init__(self, fget, default=None, **kwargs):
        self.fget = fget
        super().__init__(default=default, **kwargs)

    def get_value(self, conf):
        value = super().get_value(conf)
        return self.fget(conf, value)


class Env(environ.Env):
    def __call__(self, name, default=None, type=None, **kwargs):
        if type is None and default is None:
            type = str
        elif type is None:
            type = default.__class__

        method_name = EnvDescriptor.METHOD_MAPPER[type]
        method = getattr(self, method_name)
        return method(name, default=default, **kwargs)



import sidekick as sk

from .queryset import QuerySet

wc = sk.import_later('wordcloud')


class WordCloudQuerySet(QuerySet):
    """
    A queryset that exposes a word_cloud() method to display a word cloud from
    some fields extracted from a queryset.
    """

    def word_cloud(self, *fields, **kwargs):
        """
        Return an word cloud from the given fields spanning all values in the
        queryset.

        See Also:
            see :func:`word_cloud` for additional options.
        """
        return word_cloud(self, fields=fields, **kwargs)


def word_cloud(qs, fields=(), lang='en', sep='\n', stop_words=None, **kwargs):
    """
    Creates a word-cloud object from queryset.

    Args:
        qs:
            Queryset
        fields:
            Fields used to extract text from queryset.
        lang:
            Language in which the word cloud is defined.
        stop_words:
            List of words that should be ignored in the final result.
        sep:
            Separator used to join content of different entries of the queryset.
            Defaults to a new line.
    """
    data = sep.join(sk.flatten(qs.values_list(*fields)))
    cloud = wc.WordCloud(**kwargs)
    return cloud.generate(data)



import os

from boogie.configurations.tools import module_exists
from .paths import PathsConf


class TemplatesConf(PathsConf):
    """
    Configure templates.
    """

    def get_templates(self):
        templates = [self.DJANGO_TEMPLATES, self.JINJA_TEMPLATES]
        return [x for x in templates if x]

    def get_django_templates(self):
        return {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': self.DJANGO_TEMPLATES_DIRS,
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        }

    def get_jinja_templates(self):
        options = {}
        env = self.get_jinja2_environment()
        if env is not None:
            options['environment'] = env
        return {
            'BACKEND': 'django.template.backends.jinja2.Jinja2',
            'DIRS': self.JINJA_TEMPLATES_DIRS,
            'APP_DIRS': True,
            'OPTIONS': {
                'extensions': self.JINJA2_EXTENSIONS,
                **options,
            },
        }

    def get_django_templates_dirs(self):
        return []

    def get_jinja_templates_dirs(self):
        return []

    def get_jinja2_extensions(self):
        return ['jinja2.ext.i18n']

    def get_jinja2_environment(self):
        base, _, end = os.environ['DJANGO_SETTINGS_MODULE'].rpartition('.')
        module = base + '.jinja2'
        if module_exists(module):
            return module + '.environment'
        return 'boogie.jinja2.boogie_environment'



from boogie.configurations import django_conf
from boogie.configurations.django_conf import DjangoConf


class LoggingConf(django_conf.LoggingConf):
    """
    Same as default config.
    """


class UrlsConf(django_conf.UrlsConf):
    """
    Same as default config.
    """


class PathsConf(django_conf.PathsConf):
    """
    Same as default config.
    """


class TemplatesConf(django_conf.TemplatesConf):
    """
    Same as default config.
    """


class EnvironmentConf(django_conf.EnvironmentConf):
    """
    Same as default config.
    """


class ServicesConf(django_conf.ServicesConf):
    """
    Same as default config.
    """


class SecurityConf(django_conf.SecurityConf):
    """
    Same as default config.
    """


class InstalledAppsConf(django_conf.InstalledAppsConf):
    """
    Install the following apps:
    * Django Rest Framework apps

    Plus:
    * Install Django Debug Toolbar in local environment.
    """

    def get_third_party_apps(self):
        apps = super().get_third_party_apps()
        if self.ENVIRONMENT == 'local':
            apps = ['debug_toolbar', *apps, 'django_extensions']
        return apps


class MiddlewareConf(django_conf.MiddlewareConf):
    """
    Changes:
    * Install Django debug toolbar middleware in local enviroment.
    """


class BoogieConf(DjangoConf,
                 SecurityConf,
                 TemplatesConf,
                 UrlsConf,
                 MiddlewareConf,
                 InstalledAppsConf,
                 LoggingConf,
                 ServicesConf,
                 PathsConf,
                 EnvironmentConf):
    """
    Base configuration for the Boogie stack.

    See each layer of configuration to see the changes introduced by the
    Boogie stack.
    """



from . import locales
from .database import DatabaseConf
from .environment import EnvironmentConf
from .installed_apps import InstalledAppsConf
from .locales import LocaleConf, country
from .logging import LoggingConf
from .middleware import MiddlewareConf
from .paths import PathsConf
from .security import SecurityConf
from .services import ServicesConf
from .templates import TemplatesConf
from .urls import UrlsConf


class DjangoConf(SecurityConf,
                 TemplatesConf,
                 UrlsConf,
                 MiddlewareConf,
                 InstalledAppsConf,
                 LoggingConf,
                 LocaleConf,
                 DatabaseConf,
                 ServicesConf,
                 PathsConf,
                 EnvironmentConf):
    """
    Base configuration class for Django-based projects.
    """



import importlib.util
import os
from pathlib import Path

from .environment import EnvironmentConf


class PathsConf(EnvironmentConf):
    """
    Configure the default paths for a Django project
    """

    def get_repo_dir(self):
        """
        Return the repository directory.
        """
        # Try to guess the base repository directory. We first assume the
        # project uses git and look for a parent folder with a .git tree
        path = git_folder(self)
        if path is not None:
            return path

        # The final guess is the current directory
        return os.getcwd()

    def get_settings_file_path(self):
        django_settings = os.environ.get('DJANGO_SETTINGS_MODULE', None)
        if django_settings:
            spec = importlib.util.find_spec(django_settings)
            return Path(spec.origin)

    def get_config_dir(self):
        settings = self.SETTINGS_FILE_PATH
        if settings:
            return None
        if settings.name == '__init__.py':
            settings = settings.parent
        return settings.parent

    def get_base_dir(self):
        return get_dir(self)

    def get_log_file_path(self):
        value = self.env('DJANGO_LOG_FILE_PATH', default=None)
        if value is None:
            value = self.BASE_DIR / 'logfile.log'
        return value

    def get_django_project_path(self):
        name, _, _ = os.environ['DJANGO_SETTINGS_MODULE'].rpartition('.')
        return name


#
# Auxiliary functions
#
def git_folder(conf):
    paths = []
    if conf.SETTINGS_FILE_PATH:
        paths.append(conf.SETTINGS_FILE_PATH.parent)

    if not conf.__module__.startswith('boogie'):
        spec = importlib.util.find_spec(conf.__module__)
        paths.append(Path(spec.origin).parent)

    paths.append(Path(os.getcwd()))

    for path in paths:
        for subpath in [path, *path.parents]:
            if (subpath / '.git').exists():
                return subpath


def get_dir(conf):
    spec = importlib.util.find_spec(conf.__module__)
    return Path(spec.origin).parent



from django.conf import settings

REGISTERED_FUNCTIONS = {}

MIN_USERS = getattr(settings, 'EJ_MATH_MIN_USERS', 5)
MIN_COMMENTS = getattr(settings, 'EJ_MATH_MIN_COMMENTS', 5)
MIN_VOTES = getattr(settings, 'EJ_MATH_MIN_VOTES', 5)
USE_CELERY = bool(getattr(settings, 'CELERY_BROKER', None))



import inspect
import itertools
import logging
import os
import sys
from importlib import import_module

from django.core.exceptions import ImproperlyConfigured

from .descriptors import Env

log = logging.getLogger('boogie')
NOT_GIVEN = object()


def save_configuration(conf_class, where=None):
    """
    Prepare Django to receive the current configurations. This function is
    usually called at the end of the configuration module and it creates an
    instance of the configuration class and injects all names in its namespace.

    Usage::

        from boogie.configurations import set_configuration, Default

        class Conf(Default):
            MY_OPTION = 42

        init_configuration(Conf)


    Args:
        conf_class:
            A configuration class.
        where:
            The module that it should inject configuration parameters. It can be
            a dictionary or a string with the module path.
    """
    if where is None:
        try:
            where = os.environ['DJANGO_SETTINGS_MODULE']
        except KeyError:
            raise ImproperlyConfigured(
                'You must either define the DJANGO_SETTINGS_MODULE environment '
                'variable or pass an explicit module name to the '
                'init_configuration function.'
            )
    if isinstance(where, str):
        where = import_module(where).__dict__
    elif isinstance(where, type(sys)):
        where = where.__dict__

    # We create an instance of the class configuration and import all symbols
    # with upper case names.
    try:
        conf = conf_class()
        where.update(conf.load_settings())
    except Exception as exc:
        log.error('Error loading configurations: %s' % exc)
        raise


class Conf:
    """
    Base class for all configuration classes.
    """

    ENVFILE = None
    env_prefix = ''

    @classmethod
    def save_settings(cls, where=None):
        """
        Calls :func:`save_configuration` with the current class.
        """
        save_configuration(cls, where)

    def __init__(self, **kwargs):
        self._settings = None
        self.env = Env()

        cls = type(self)
        for name, value in kwargs.items():
            attr = name.upper()
            if hasattr(cls, attr) or hasattr(cls, f'get_{name}'):
                setattr(self, attr, value)
            else:
                raise TypeError(f'invalid argument: {attr}')

    def prepare(self):
        """
        A hook that is executed once after initialization and before creating
        the default settings dictionary.

        The default implementation tries to read an env file if it is defined
        by the ENVFILE settings variable.
        """
        if self.ENVFILE:
            self.env.read_env(self.ENVFILE)

    def finalize(self, settings):
        """
        A hook that receives a settings dictionary and returns the final
        output of the get_settings() method.

        The default implementation simply pass the input dictionary forward.
        """
        return settings

    def load_settings(self):
        """
        Return a dictionary with all settings defined by the configuration.

        This method is idempotent. It caches the resulting settings and always
        return the same settings dictionary.
        """
        if self._settings is None:
            self.prepare()

            # Load standard settings
            settings = {
                attr: getattr(self, attr)
                for attr in set(with_getter_attributes(dir(self)))
            }
            settings = {k: v for k, v in settings.items() if v is not NOT_GIVEN}
            if settings.get('ENVFILE') is None:
                settings.pop('ENVFILE', None)
            self._settings = self.finalize(settings)
        return dict(self._settings)

    def __getattr__(self, attr):
        if attr.isupper():
            func = getattr(self, f'get_{attr.lower()}', None)
            if func is None:
                raise AttributeError(f'Invalid attribute {attr}')
            value = get_value(func, self, attr)
            setattr(self, attr, value)
            return value
        raise AttributeError(attr)


#
# Auxiliary methods
#
def get_value(func, ns, which):
    """
    Evaluate function using given namespace to inject arguments.
    """
    spec = inspect.getfullargspec(func)
    callargs = {}

    with_default = itertools.chain(
        args_with_default(spec.args, spec.defaults or (), NOT_GIVEN),
        [(k, spec.kwonlydefaults[k]) for k in spec.kwonlyargs],
    )
    for name, default in with_default:
        # Skip self
        if name == 'self':
            continue

        # The "env" parameter injects the environment variable associated with
        # the attribute
        if name == 'env':
            prefix = ns.env_prefix
            attr = which.upper()
            var_name = getattr(func, 'env_name', f'{prefix}{attr}')
            type = getattr(func, 'env_type', str)
            default = getattr(func, 'env_default', default)
            value = ns.env(var_name, type=type, default=default)
            if value is NOT_GIVEN:
                name = which

        # Otherwise we just fetch the variable from the given namespace
        else:
            value = getattr(ns, name.upper(), default)

        # Save variables in dictionary
        if value is NOT_GIVEN:
            var_name = name.upper()
            msg = f'{which}: configuration must define a {var_name} attribute'
            raise TypeError(msg)
        callargs[name] = value

    return func(**callargs)


def args_with_default(names, defaults, fillvalue=None):
    """
    Iterate over pairs of (argument, default) values.
    """
    rnames = reversed(names)
    rdefaults = reversed(defaults)
    pairs = itertools.zip_longest(rnames, rdefaults, fillvalue=fillvalue)
    return reversed(list(pairs))


def with_getter_attributes(attrs):
    for attr in attrs:
        if attr.isupper() and not attr.startswith('_'):
            yield attr
        elif attr.startswith('get_'):
            attr = attr[4:].upper()
            yield attr



from functools import wraps

from django.db import models

#
# Global constants
#
QUERYSET_METHODS = {}
MANAGER_METHODS = {}


#
# Public decorators and functions
#
def manager_method(model, name=None):
    """
    Decorator that marks function to be injected as a manager method for the
    supplied model.
    """
    return _manager_method(model, name)


def queryset_method(model, skip_manager=False, name=None):
    """
    Decorator that marks function to be injected as a queryset method for the
    supplied model.
    """

    def decorator(func):
        if not skip_manager:
            manager_func = (lambda m: func(m.get_queryset()))
            manager_method(model, name=name)(wraps(func)(manager_func))
        return _queryset_method(model, name=name)(func)

    return decorator


def get_queryset_attribute(qs, attr):
    """
    Return an attribute registered with the :func:`queryset_method` decorator.
    """
    return get_attribute(qs, attr, QUERYSET_METHODS)


def get_manager_attribute(qs, attr):
    """
    Return an attribute registered with the :func:`manager_method` decorator.
    """
    return get_attribute(qs, attr, MANAGER_METHODS)


#
# Auxiliary functions
#
def get_attribute(qs, attr, registry):
    """
    Common implementation to `func`:get_queryset_attr
    """

    if attr.startswith('_'):
        raise AttributeError(attr)

    model = qs.model
    try:
        return registry[(model, attr)](qs)
    except KeyError:
        pass

    # Search MRO
    django_model = models.Model
    for cls in model.mro():
        if issubclass(cls, django_model):
            try:
                method = registry[(cls, attr)]
            except KeyError:
                continue
            registry[(model, attr)] = method
            return get_descriptor(method, qs)
    else:
        raise AttributeError(attr)


def get_descriptor(descriptor, instance):
    try:
        getter = descriptor.__get__
    except AttributeError:
        return descriptor
    else:
        return getter(instance, type(instance))


def registry_decorator(registry):
    """
    Common implementation to @queryset and @manager decorators.
    """

    def register(model, name=None):
        def decorator(obj):
            attr = name or obj.__name__
            if hasattr(obj, '__get__'):
                getter = obj.__get__
                registry[(model, attr)] = lambda qs: getter(qs, type(qs))
            else:
                registry[(model, attr)] = lambda qs: obj
            return obj

        return decorator

    return register


_queryset_method = registry_decorator(QUERYSET_METHODS)
_manager_method = registry_decorator(MANAGER_METHODS)



import random
import string

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import reverse
from django.utils import translation
from jinja2 import Environment
from markdown import markdown
from markupsafe import Markup

SALT_CHARS = string.ascii_letters + string.digits + '-_'


def environment(**options):
    options.pop('debug', None)
    options.setdefault('trim_blocks', True)
    options.setdefault('lstrip_blocks', True)
    env = Environment(**options)
    env.globals.update(
        static=staticfiles_storage.url,
        url=reverse,
        salt_attr=salt_attr,
        salt_tag=salt_tag,
        salt=salt,
        service_worker=getattr(settings, 'SERVICE_WORKER', False),
    )
    env.filters.update(
        markdown=lambda x: Markup(markdown(x)),
        salt=salt,
    )
    env.install_gettext_translations(translation, newstyle=True)
    return env


#
# Security
#
def salted(value):
    """
    Protects a value using a random salt. This completely prevents the BREACH
    attack against a salted secret, but stores values in a codified form that
    cannot be directly used.

    The BREACH authors suggest a transform that takes a string P and returns
    (P + (S^P)), where S^P is a XOR operation of each byte. Templates do not
    have a byte-level control of the message, hence salting must be done in
    unicode data. This brings some problems:

    * How to (efficiently) implement XOR for a large alphabet such as unicode?
    * Xor-ing two valid unicode points will not necessarily yield a valid
      unicode value.
    * Client code needs to recover the encoded data, hence the solution must
      be portable to Python and Javascript
    """
    # TODO: how to do it with unicode?
    raise NotImplementedError


def salt(size=None):
    """
    Create a random string of characters of a random size.

    The goal of the random salt is to make it more difficult to execute a
    BREACH attack (https://docs.djangoproject.com/en/2.0/ref/middleware/#module-django.middleware.gzip).
    A random salt makes the attack more difficult by requiring more requests to
    correctly guess each byte in the target secret sub-string.
    """
    size = random.randint(4, 10) if size is None else size
    func = random.choice
    chars = SALT_CHARS
    return ''.join(func(chars) for _ in range(size))


def salt_attr():
    """
    A salt added as an HTML attribute.
    """
    return f'data-salt="{salt()}"'


def salt_tag():
    """
    A salt added as an invisible HTML tag.
    """
    return f'<div style="display: none" {salt_attr()}></div>'



from .environment import EnvironmentConf
from ..descriptors import env


def using_db(db):
    return lambda self: db in self.DATABASES['default']['ENGINE']


class DatabaseConf(EnvironmentConf):
    """
    Configure the database.

    See also: https://docs.djangoproject.com/en/2.0/ref/settings/#databases
    """

    DATABASE_FROM_DB_URL = env('sqlite:///local/db/db.sqlite3',
                               type='db_url',
                               name='DJANGO_DB_URL')

    def get_databases(self):
        return {'default': {'TEST': {}, **self.DATABASE_FROM_DB_URL}}

    # Derived inspections
    get_using_sqlite = using_db('sqlite')
    get_using_postgresql = using_db('postgresql')
    get_using_mysql = using_db('mysql')



from rest_framework import viewsets, status
from rest_framework.response import Response


class RestAPIBaseViewSet(viewsets.ModelViewSet):
    """
    Standard boogie view set.
    """

    queryset = None
    serializer_class = None
    lookup_field = None
    base_name = None
    api_version = None

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        try:
            self.perform_destroy(instance)
            return Response(status=status.HTTP_204_NO_CONTENT)
        except PermissionError as exc:
            return wrap_permission_error(exc)

    def perform_destroy(self, instance):
        self.delete_hook(self.request, instance)

    def delete_hook(self, request, instance):
        instance.delete()

    def get_queryset(self):
        return self.query_hook(self.request, self.queryset).all()

    def query_hook(self, request, qs):
        return qs


def wrap_permission_error(error):
    return Response({
        'error': 'permission',
        'code': 403,
        'message': str(error),
    })



from copy import deepcopy

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as AuthUserAdmin

from . import models


@admin.register(models.User)
class UserAdmin(AuthUserAdmin):
    # Remove "first_name" and "last_name" and replace it by "name"
    fieldsets = list(deepcopy(AuthUserAdmin.fieldsets))
    fieldsets[1][1]['fields'] = ('name', 'email')
    fieldsets = tuple(fieldsets)



from .descriptors import env, env_settings, env_default
from .base import Conf, save_configuration
from .django_conf import (
    DjangoConf, EnvironmentConf, InstalledAppsConf, MiddlewareConf, TemplatesConf,
    SecurityConf, PathsConf, UrlsConf, ServicesConf, LoggingConf, locales,
)



from functools import wraps

from django.db.models import Manager
from sidekick import lazy, placeholder as this


def manager_only(method=None):
    """
    Decorator that marks a method as a manager method, i.e., it can only be
    accessed by a manager created with QuerySet.as_manager() and not by
    instances of the queryset itself.

    Usage:

        .. code-block:: python

            # managers.py

            class BookQuerySet(models.QuerySet):
                @manager_only()
                def new_edition(self, book, edition):
                    new = copy(book)
                    new.edition = edition
                    new.save()
                    return new

            BookManager = Manager.from_queryset(BookQuerySet)
    """

    if method is None:
        return manager_only

    name = method.__name__
    msg = 'this method cannot be used from a queryset.'
    if name != '<lambda>':
        msg = '{name} method cannot be used from a queryset.'

    @wraps(method)
    def decorated(self, *args, **kwargs):
        if isinstance(self, Manager):
            return method(self, *args, **kwargs)
        else:
            raise TypeError(msg.format(name=name))

    return decorated


def manager_property(method=None):
    """
    Decorates a method that returns a QuerySet instance and expose it as a
    manager. This is useful to define custom accessors for related value,
    providing a consistent api between your custom managers and the related
    managers automatically created by Django.

    Usage:

        .. code-block:: python

            # models.py

            class Book(models.Model):
                authors = models.ManyToManyField('User')
                ...

                @manager_property()
                def active_authors(self):
                    return self.authors.filter(active=True)


        The ``active_authors`` property is a manager, not a queryset. While most
        manager methods are available to query sets, making them interchangeable
        in many situations, exposing the property as actual managers
        improves the API consistency.

        >>> [author.name for author in book.active_authors.all()]
        ['Douglas Adams', ...]
    """

    if method is None:
        return manager_property

    @wraps(method)
    def wrapped(self):
        return RelatedManager(self, method(self))

    return property(wrapped)


#
# Auxiliary classes
#
def alters_data(func):
    func.alters_data = True
    return func


class RelatedManager(Manager):
    manager = lazy(this.queryset)
    model = lazy(this.queryset.model)
    instance_field = lazy(this.queryset.field.name)
    do_not_call_in_templates = True

    def __init__(self, instance, queryset, instance_field=None):
        super().__init__()
        self.instance = instance
        self.queryset = queryset
        if instance_field is not None:
            self.instance_field = instance_field

    def __getattr__(self, item):
        return getattr(self.manager, item)

    def get_queryset(self):
        return self.queryset

    @alters_data
    def add(self, *objs, bulk=True):
        return self.manager.add(*objs, bulk=True)

    @alters_data
    def create(self, **kwargs):
        kwargs.update(self._construct_params())
        return self.manager.create(**kwargs)

    @alters_data
    def get_or_create(self, **kwargs):
        kwargs.update(self._construct_params())
        return self.manager.get_or_create(**kwargs)

    @alters_data
    def update_or_create(self, **kwargs):
        kwargs.update(self._construct_params())
        return self.manager.update_or_create(**kwargs)

    # remove() and clear() are only provided if the ForeignKey can have a
    # value of null.
    @alters_data
    def remove(self, *objs, bulk=True):
        self._check_nullable_relation()
        return self.manager.remove(*objs, bulk=True)

    @alters_data
    def clear(self, *, bulk=True):
        self._check_nullable_relation()
        return self.manager.clear(bulk=True)

    @alters_data
    def set(self, objs, *, bulk=True, clear=False):
        return self.manager.set(objs, bulk=True, clear=False)

    #
    # Auxiliary methods
    #
    def _check_nullable_relation(self):
        ...  # TODO: check if foreign key relation can hold null values.

    def _construct_params(self):
        # TODO: return mapping with additional params that must be set by
        # using the inverse relation when creating new objects.
        return {}



import logging

from django.db.models import fields, Model
from django.urls import register_converter
from django.urls.converters import get_converter

log = logging.getLogger('boogie')

FIELD_TYPES_MAP = {
    # Numeric
    fields.IntegerField: 'int',
    fields.SmallIntegerField: 'int',
    fields.PositiveSmallIntegerField: 'int',

    # UUID
    fields.UUIDField: 'uuid',

    # Slugs
    fields.SlugField: 'slug',
}


class ModelConverterBase:
    """
    Specialized converter for specific models.

    By default, it will register the

    <app_label.model_name.lookup_value:value>
    """

    model = None
    regex = None
    base_converter = None
    raw_converter = None
    lookup_field = 'pk'

    def to_python(self, path):
        if isinstance(path, Model):
            return path
        value = self.base_converter.to_python(path)
        query = {self.lookup_field: value}
        try:
            return self.model.objects.get(**query)
        except self.model.DoesNotExist:
            return None

    def to_url(self, model):
        if isinstance(model, str):
            return model
        return getattr(model, self.lookup_field)


def register_model_converter(model, type_name, lookup_field='pk', lookup_type=None):
    """
    Register a converter for the given model.

    Args:
        model:
            A Django model class.
        type_name:
            Converter type name. The type name cannot include colons or angle
            brackets, but something like "auth.user" is ok. This is the
            name used in path expressions to identify the converter like in
            path("profiles/<auth.user:user>", view, ...)
        lookup_field:
            Lookup field used to extract the model instance from the database.
            (defaults to primary key).
        lookup_type:
            Optional converter type of the lookup field. Can be useful for
            things like slugs, but usually can be derived from the lookup_field.
    """
    class_name = f'{model.__name__}Converter'
    raw_converter = get_lookup_type(lookup_type, model, lookup_field)
    base_converter = get_converter(raw_converter)
    converter = type(class_name, (ModelConverterBase,), {
        'model': model,
        'lookup_field': lookup_field,
        'raw_converter': raw_converter,
        'base_converter': base_converter,
        'regex': base_converter.regex,
    })
    register_converter(converter, type_name)


def get_lookup_type(kind, model, field_name):
    if kind is not None:
        return kind

    if field_name == 'pk':
        field = model._meta.pk
    else:
        field = model._meta.get_field(field_name)
    return FIELD_TYPES_MAP.get(type(field), 'str')



from functools import lru_cache

from django.db.models import BooleanField
from django.db.models.expressions import Combinable


def _method(op):
    return lambda self, other: self._boolean_combine(other, op)


class Comparable(Combinable):
    """
    Mixin class that adds comparison operators (==, != , >, <, >=, <=) to the
    ``Combinable`` interface.
    """

    # Comparison connectors
    GT = '>'
    GTE = '>='
    LT = '<'
    LTE = '<='
    EQ = '='
    NEQ = '<>'

    __gt__ = _method(GT)
    __lt__ = _method(LT)
    __ge__ = _method(GTE)
    __le__ = _method(LTE)
    __eq__ = _method(EQ)
    __ne__ = _method(NEQ)

    def _boolean_combine(self, other, op):
        expression = self._combine(other, op, False)
        expression.output_field = BooleanField()
        expression.comparable_expression = True
        return expression

    def equal(self, other):
        return super().__eq__(other)

    def not_equal(self, other):
        return super().__ne__(other)


def as_comparable(obj):
    """
    Return object as an instance of Comparable class.
    """
    cls = get_comparable_class(type(obj))
    obj.__class__ = cls
    return obj


@lru_cache(maxsize=256)
def get_comparable_class(cls, doc=None):
    if issubclass(cls, Comparable):
        return cls
    ns = {'__doc__': doc} if doc else {}
    return type(cls.__name__ + 'Comparable', (Comparable, cls), ns)



import collections
from functools import lru_cache, partial, singledispatch

from django.db import models
from django.db.models.query import ValuesListIterable
from sidekick import lazy, itertools, import_later, alias

from boogie.models.methodregistry import get_queryset_attribute
from boogie.models.utils import LazyMethod
from .expressions import F
from ..utils.linear_namespace import linear_namespace

F_EXPR_TYPE = type(F.some_attr > 0)

# LAZY IMPORTS
# Useful queryset methods
# import bulk_update.manager
# import manager_utils
pd = import_later('pandas')


# noinspection PyProtectedMember
class QuerySet(models.QuerySet):
    """
    Boogie's drop in replacement to Django's query sets.

    It extends the query set API with a Pydata-inspired interface to select
    data.
    """

    # Manager utils
    id_dict = LazyMethod('manager_utils:ManagerUtilsMixin.id_dict')
    bulk_upsert = LazyMethod('manager_utils:ManagerUtilsMixin.bulk_upsert')
    sync = LazyMethod('manager_utils:ManagerUtilsMixin.sync')
    upsert = LazyMethod('manager_utils:ManagerUtilsMixin.upsert')
    get_or_none = LazyMethod('manager_utils:ManagerUtilsMixin.get_or_none')
    single = LazyMethod('manager_utils:ManagerUtilsMixin.single')

    # Bulk update from manager utils is very limited, we use the implementation
    # in the django-bulk-update package.
    bulk_update = LazyMethod('bulk_update.manager:BulkUpdateManager.bulk_update')

    # Properties
    index = property(lambda self: Index(self))
    _selected_column_names = None

    # Class methods
    def as_manager(cls):
        # Overrides to use Boogie manager instead of the default manager
        from .manager import Manager
        manager = Manager.from_queryset(cls)()
        manager._built_with_as_manager = True
        return manager

    as_manager.queryset_only = True
    as_manager = classmethod(as_manager)

    new = alias('model')

    def __getitem__(self, item):
        # 1D indexing is delegated to Django. We extend it with some
        # idioms that preserve backwards compatibility
        #
        # We keep Django's behavior in all of the following cases (a, b >= 0):
        # * qs[a]
        # * qs[a:b]
        #
        # Small extensions:
        # * qs[-a] -> fetch from last
        # * qs[0:-b] -> specify a index from first to last
        #
        # Larger extensions:
        # * qs[set] -> fetch elements in the given pk set
        # * qs[list] -> fetch specified elements by pk in order
        return get_queryset_item(item, self)

    def __setitem__(self, item, value):
        # Similarly to getitem, we dispatch for the 1d and 2d indexing
        # functions
        if not isinstance(item, tuple):
            return setitem_1d(self, item, value)

        try:
            row, col = item
        except IndexError:
            raise TypeError('only 1d or 2d indexing is allowed')

        return setitem_2d(self, row, col, value)

    def __getattr__(self, attr):
        return get_queryset_attribute(self, attr)

    # Better accessors
    def update_item(self, pk, **kwargs):
        """
        Updates a single item in the queryset.
        """
        return self.filter(pk=pk).update(**kwargs)

    update_item.alters_data = True

    #
    # Enhanced API
    #
    def select_columns(self, *fields):
        """
        Similar to .values_list(*fields)
        """
        return select_columns(self, list(fields))

    def annotate_verbose(self, **fields):
        """
        Annotate given fields with their verbose versions.
        """
        raise NotImplementedError

    def auto_annotate_verbose(self, *fields):
        raise NotImplementedError

    def values(self, *fields, verbose=False, **expressions):
        if verbose:
            return self.auto_annotate_verbose(fields)

        qs = super().values(*fields, **expressions)
        fields = fields + tuple(expressions.keys())
        return self._mark_column_names(fields, qs)

    def values_list(self, *fields, verbose=False, **kwargs):
        if verbose:
            return self.auto_annotate_verbose(fields)
        qs = super().values_list(*fields, **kwargs)
        return self._mark_column_names(fields, qs)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        if args and getattr(args[0], 'comparable_expression', False):
            expr, *args = args
            name = f'filter_{id(expr)}'
            kwargs[name] = True
            qs = self.annotate(**{name: expr})
            # noinspection PyProtectedMember
            return qs._filter_or_exclude(negate, *args, **kwargs)
        return super()._filter_or_exclude(negate, *args, **kwargs)

    def _mark_column_names(self, columns, qs=None):
        qs = self if qs is None else qs
        qs._selected_column_names = columns
        return qs

    #
    # Pandas data frame and numpy array APIs
    #
    def dataframe(self, *fields, index=None, verbose=False):
        """
        Convert query set to a Pandas data frame.

        If fields are given, it uses a similar semantics as .values_list(),
        otherwise, it uses the selected fields or the complete set of fields.

        Args:
            index:
                Name of index column (defaults to primary key).
            verbose (bool):
                If given, prints foreign keys ad choices using human readable
                names.
        """
        if not fields:
            if self._selected_column_names:
                fields = self._selected_column_names
            else:
                fields = [f.name for f in self.model._meta.fields
                          if index is None and not f.primary_key]
        elif len(fields) == 1 and isinstance(fields[0], collections.Mapping):
            field_map = fields[0]
            df = self.dataframe(*field_map.values(), index=index, verbose=verbose)
            df.columns = field_map.keys()
            return df

        # Build data frame
        if index is None:
            index = self.model._meta.pk.name

        data = list(self.values_list(index, *fields, verbose=verbose))
        df = pd.DataFrame(data, columns=['__index__', *fields])
        df.index = df.pop('__index__')
        df.index.name = index
        return df

    def pivot_table(self, index, columns, values,
                    verbose=False, dropna=False, fill_value=None):
        """
        Creates a pivot table from this queryset.

        Args:
            index:
                Field used to define the pivot table indexes (rows names).
            columns:
                Field used to populate the different columns.
            values:
                Field used to fill table with values.
            dropna (bool):
                If True (default), exclude columns whose entries are all NaN.
            fill_value:
                Value to replace missing values with.
            verbose (bool):
                If given, prints foreign keys ad choices using human readable
                names.

        """
        df = self.dataframe(index, columns, values, verbose=verbose)
        return df.pivot_table(index=index, columns=columns, values=values,
                              dropna=dropna, fill_value=fill_value)

    def update_from_dataframe(self, dataframe, batch_size=None, in_bulk=True):
        """
        Persist data frame data to the database. Data frame index must
        correspond to primary keys of existing objects.

        Args:
            dataframe:
                A pandas data frame
            in_bulk (bool):
                If True (default), save values in bulk.
            batch_size:
                If saving in bulk, defines the size of each batch that touches
                the database. This avoids creating very long SQL commands that
                can halt the database for a perceptible amount of time.
        """
        objects = []
        fields = dataframe.columns
        add_object = objects.append
        new_object = self.model

        for pk, row in zip(dataframe.index, dataframe.to_dict('records')):
            row.setdefault('pk', pk)
            add_object(new_object(**row))

        if in_bulk:
            self.bulk_update(objects,
                             update_fields=list(fields),
                             batch_size=batch_size)
        else:
            for obj in objects:
                obj.save(update_fields=fields)

    def extend_dataframe(self, df, *fields, verbose=False):
        """
        Returns a copy of dataframe that includes columns computed from the
        given fields.
        """
        return df.append(
            self.filter(pk__in=df.index)
                .dataframe(*fields, verbose=verbose))

    #
    # Selecting parts of the dataframe
    #
    def head(self, n=5):
        """
        Select the first n rows in the query set.
        """
        return self[:n]

    def tail(self, n=5):
        """
        Select the last n rows in the query set.
        """
        return self[-n:]


#
# Auxiliary classes
# ----------------------------------------------------------------------------
class QuerysetSequence(collections.Sequence):
    @staticmethod
    def prepare(queryset):
        return queryset

    _prepared = lazy(lambda self: self.prepare(self.queryset))

    def __init__(self, queryset):
        self.queryset = queryset

    def __iter__(self):
        return iter(self._prepared)

    def __len__(self):
        return len(self._prepared)

    def __getitem__(self, item):
        return self._prepared[item]

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, list(self))


class Index(QuerysetSequence):
    """
    Returns an index with the primary keys of all elements.
    """

    @staticmethod
    def prepare(qs):
        return qs.values_list('pk', flat=True)


class RowIterable(ValuesListIterable):
    """
    Iterable returned by QuerySet.values_list(named=True) that yields a
    namedtuple for each row.
    """

    @staticmethod
    @lru_cache()
    def _create_row_class(names):
        return linear_namespace('Row', names)

    def __init__(self, *args, **kwargs):
        self.field_names = kwargs.pop('field_names', None)
        super().__init__(*args, **kwargs)

    def __iter__(self):
        queryset = self.queryset
        if self.field_names:
            names = self.field_names
        elif queryset._fields:
            names = queryset._fields
        else:
            query = queryset.query
            names = list(query.extra_select)
            names.extend(query.values_select)
            names.extend(query.annotation_select)

        new = self._create_row_class(tuple(names))
        for row in super().__iter__():
            yield new(*row)


#
# Getitem auxiliary functions
# ----------------------------------------------------------------------------

def getitem_2d(qs, rows, cols):
    """
    Extract elements from queryset possibly filtering specific rows
    or columns.

    The first index selects rows and the second columns.
    """
    # First we filter the rows
    if is_vector_row_access(rows):
        qs = filter_queryset(qs, rows)
        return select_columns(qs, cols)
    else:
        qs = qs.filter(pk=rows)
        return select_columns(qs, cols)[0]


def is_vector_row_access(row):
    """
    Return True to all types that select multiple rows.
    """
    return isinstance(row, (list, slice, QuerySet, F_EXPR_TYPE))


def filter_queryset(qs, rows):
    """
    Extended ways to filter a queryset.
    """

    # Filter by pk
    if isinstance(rows, list):
        model = qs.model
        pks = {x.pk if isinstance(x, model) else x for x in rows}
        return qs.filter(pk__in=pks)

    elif isinstance(rows, slice):
        if rows == slice(None, None, None):
            return qs
        raise IndexError('invalid column slice: %s' % rows)

    elif isinstance(rows, QuerySet):
        return qs & rows

    elif isinstance(rows, F_EXPR_TYPE):
        return qs.filter(rows)

    raise TypeError('invalid row selector: %r' % rows.__class__.__name__)


def select_columns(qs, cols):
    if isinstance(cols, list):
        qs = qs.values_list(*cols, flat=False)
        field_names = get_column_names(cols, qs)
        qs._iterable_class = partial(RowIterable, field_names=field_names)
        return qs

    elif isinstance(cols, slice):
        if cols == slice(None, None, None):
            return qs
        raise IndexError('invalid column slice: %s' % cols)

    elif isinstance(cols, (str, F, models.F)):
        return qs.values_list(cols, flat=True)

    raise TypeError('invalid column selector: %r' % cols.__class__.__name__)


def get_column_names(cols, qs):
    """
    Return a list of valid field names for the given columns.
    """
    names = []
    append = names.append
    for col in cols:
        if isinstance(col, str):
            name = col
        elif isinstance(col, F):
            name = col._name
        else:
            raise TypeError('invalid column', col)
        append(name)
    return names


#
# Setitem auxiliary functions
# ----------------------------------------------------------------------------
@singledispatch
def get_queryset_item(item, qs):
    raise TypeError(f'Invalid index type: {item.__class__.__name__}')


@get_queryset_item.register(tuple)
def _(item, qs):
    try:
        row, col = item
    except IndexError:
        raise TypeError('only 1d or 2d indexing is allowed')
    return getitem_2d(qs, row, col)


@get_queryset_item.register(int)
def _(n, qs):
    if n >= 0:
        return super(QuerySet, qs).__getitem__(n)
    else:
        rev = qs.reverse() if qs.ordered else qs.order_by('-id')
        return rev[abs(n) - 1]


@get_queryset_item.register(slice)
def _(slice, qs):
    start, stop, step = slice.start, slice.stop, slice.step
    if step is None and \
            (start is None or start >= 0) and \
            (stop is None or stop >= 0):
        return super(QuerySet, qs).__getitem__(slice)
    elif stop <= 0:
        if start != 0 and start is not None:
            raise ValueError(
                'negative index are only allowed if starting from the '
                'beginning of the queryset')
        return qs.reverse()[abs(stop) - 1:].reverse()


@get_queryset_item.register(set)
def _(set_index, qs):
    return qs.filter(pk__in=set_index)


@get_queryset_item.register(list)
def _(lst, qs):
    raise NotImplementedError('use set instead of list')


def setitem_1d(qs, item, value):
    raise TypeError('invalid index type: %r' % item.__class__.__name__)


def setitem_2d(qs: QuerySet, rows, cols, value):
    if is_vector_row_access(rows):
        raise NotImplementedError
    elif isinstance(cols, str):
        return qs.update_item(rows, **{cols: value})
    elif isinstance(cols, list):
        cols = map(as_col_name, cols)
        mapping = dict(zip(cols, value))
        return qs.update_item(rows, **mapping)
    raise NotImplementedError


def set_columns(qs: QuerySet, cols, value):
    raise TypeError('invalid column selector: %r' % cols.__class__.__name__)


def as_col_name(obj):
    """Coerce object to a column name"""

    if isinstance(obj, str):
        return obj
    elif isinstance(obj, F):
        print(obj, obj.__dict__)
        raise ValueError('invalid column object: %s' % obj)
    return obj.name


def split_batch(seq, size, batches):
    start = 0
    it = iter(seq)
    for _ in range(batches):
        yield from itertools.islice(it, start, start + size)
        start += size



from django.contrib.contenttypes.models import ContentType
from django.db import models

from .tasks import get_functions


class TaskQuerySet(models.QuerySet):

    def last_task(self, function):
        """
        Return the last task for function
        """
        return self.filter(function=function).order_by('created').last()

    def clean_tasks(self, function, keep_last=True):
        """
        Remove old tasks for the given function.
        """
        qs = self.filter(function=function)
        if keep_last:
            qs.exclude(id=self.last_task(function).id)
        return qs.delete()

    def clean_all_tasks(self, keep_last=True):
        """
        Clean all tasks for all functions.
        """
        if keep_last:
            cls = self.model.get_object_model()
            last_tasks = [self.last_task(f).id for f in get_functions(cls)]
            return self.exclude(id__in=last_tasks).delete()
        else:
            return self.all().delete()


class TaskManager(TaskQuerySet.as_manager()):
    def create_delayed(self, function, obj, *args, **kwargs):
        """
        Creates a new promise and return it.

        Return a promise.
        """
        if self.model.get_object_model() == ContentType:
            kwargs['object_id'] = obj.id
            obj = ContentType.objects.get_for_model(type(obj))
        arguments = {'args': args, 'kwargs': kwargs}
        create = self.model._create
        return create(function=function, arguments=arguments, object=obj)

    def create_scheduled(self, function, obj, *args, **kwargs):
        """
        Similar to call(), but schedule job to run in the background using
        celery.
        """
        task = self.create_delayed(function, obj, *args, **kwargs)
        task.run(is_async=True)
        return task

    def call(self, function, obj, *args, **kwargs):
        """
        Creates a new promise and immediately execute it.

        Return the result synchronously.
        """
        task = self.create_delayed(function, obj, *args, **kwargs)
        return task.run(is_async=False)



from django.core.exceptions import ImproperlyConfigured

from ..base import Conf
from ..descriptors import env_settings, env_default


class EnvironmentConf(Conf):
    """
    Base class for Django configurations. Users should use DjangoConf
    """

    env_prefix = 'DJANGO_'

    def get_environment(self, env='local'):
        """
        Django base environment. We suggest distinguishing between 'local',
        'test' and 'production'.
        """
        if env not in ['test', 'production', 'local']:
            raise ImproperlyConfigured(f'Invalid environment: {env}')
        return env

    @env_settings(type=bool, default=None)
    def get_debug(self, env):
        """
        By default, debug is enabled only on 'local' and 'test' environments.
        """
        if env is None:
            return self.ENVIRONMENT == 'local'
        return env

    @env_default()
    def get_wsgi_application(self):
        return self.DJANGO_PROJECT_PATH + '.wsgi.application'

    @env_default()
    def get_site_id(self):
        """
        Site admin url, if Django admin app is installed.
        """
        return 1

    def get_serve_static_files(self):
        """
        It is often convenient to enable/disable Django ability to serve static
        files.
        """
        return True

    def get_auth_user_model(self):
        return 'auth.User'



from collections import Mapping

from django.core.exceptions import ImproperlyConfigured
from sidekick import lazy

from boogie.rest.utils import as_model, with_model_cache


class ApiInfo(Mapping):
    """
    Stores information about all resources of an specific API version.

    It stores the API version in an attribute and behaves as a mapping from
    resource to their respective options.
    """

    @lazy
    def resources(self):
        return list(self.values())

    def __init__(self, rest_api, version=None, parent=None):
        self.rest_api = rest_api
        self.version = version
        self.parent = parent
        self.registry = {}
        self.inline_models = {}
        self.explicit_viewsets = {}
        self.serializer_class_cache = {}

    def __getitem__(self, model):
        model = as_model(model)
        result = (
                self.registry.get(model)
                or self.inline_models.get(model)
                or self.parent and self.parent.get(model)
        )
        if result is None:
            raise KeyError(model)
        return result

    def __setitem__(self, model, value):
        model = as_model(model)
        if model in self.registry:
            raise KeyError(f'Model {model.__name__} already registered')
        self.registry[model] = value

    def __len__(self):
        n = 0 if self.parent is None else len(self.parent)
        return n + len(self.registry)

    def __iter__(self):
        yield from iter(self.registry)
        if self.parent:
            yield from self.parent

    def __repr__(self):
        data = {
            model.__name__: info.base_url for model, info in self.registry.items()
        }
        return '<ApiInfo version=%r, %r>' % (self.version, data)

    def add_hook(self, model, hook, function):
        """
        Register function to the given hook.
        """
        self[model].add_hook(hook, function)

    def base_name(self, model):
        """
        Return the base_name string for the given model.

        base_name is used to name views when the router is constructed.
        """
        return self[model].full_base_name(self.version)

    def base_url(self, model):
        """
        Return the base_name string for the given model.

        base_name is used to name views when the router is constructed.
        """
        return self[model].base_url

    def lookup_field(self, model):
        """
        Return the default lookup_field used for the given model.
        """
        return self[model].lookup_field

    def iter_viewset_items(self):
        """
        Iterates over all tuples of (base_url, viewset) for all models in this
        ApiInfo object.

        This iterator includes classes defined explicitly using register_viewset
        and implicitly by registering models with rest_api.
        """
        # Explicit viewsets
        for url, viewset in self.explicit_viewsets.items():
            yield (url, viewset)

        # Model viewsets
        for model in self:
            viewset = self.viewset_class(model)
            url = self.base_url(model)
            if url not in self.explicit_viewsets:
                yield (url, viewset)

    #
    # Registry
    #
    def register_viewset(self, base_url, viewset):
        """
        Manually associates viewset with the given url.
        """
        self.explicit_viewsets[base_url] = viewset

    def register_serializer(self, model, serializer):
        """
        Manually associates viewset with the given url.
        """
        model = as_model(model)
        self.serializer_class_cache[model] = serializer

    def register_resource(self, model, info, inline=False):
        """
        Register resource info object for model.
        """
        if inline:
            self.inline_models[model] = info
        else:
            self[model] = info

    #
    # Viewset class builder
    #
    def extra_kwargs(self, model):
        """
        Generate extra_kwargs options for the given model.
        """
        extra_kwargs = {}
        info = self[model]

        for name, related in info.related_models:
            try:
                resource_info = self[related]
            except KeyError:
                raise ImproperlyConfigured(
                    f'{model.__name__} references a {related.__name__} field as '
                    f'{name}. This model is not registered on the api and '
                    f'therefore cannot be included as reference.'
                )
            extra = resource_info.extra_kwargs(self.version)
            extra_kwargs[name] = extra
        return extra_kwargs

    @with_model_cache
    def viewset_class(self, model):
        """
        Return a viewset class for the given model.
        """

        info = self[model]
        extra = self.extra_kwargs(model)

        bases = as_bases(info.viewset_base)
        name = info.model.__name__ + 'ViewSet'
        base_name = self.base_name(model)

        namespace = {
            'Meta': viewset_meta(info, extra),
            'queryset': info.queryset,
            'serializer_class': self.serializer_class(model),
            'lookup_field': info.lookup_field,
            'base_name': base_name,
            'api_version': self.version,
            **info.action_methods,
            **info.viewset_hook_methods,
        }

        return type(name, bases, namespace)

    @with_model_cache
    def serializer_class(self, model):
        """
        Return the serializer class for the given model.
        """

        info = self[model]
        extra = self.extra_kwargs(model)

        bases = as_bases(info.serializer_base)
        name = info.model.__name__ + 'Serializer'
        base_name = self.base_name(model)

        namespace = {
            'Meta': serializer_meta(info, extra),
            'base_name': base_name,
            'detail_url': base_name + '-detail',
            'list_url': base_name + '-list',
            'lookup_field': info.lookup_field,
            'actions': list(info.detail_actions),
            'api_version': self.version,
            'explicit_links': tuple(info.links.items()),
            **info.property_methods,
            **info.serializer_hook_methods,
        }

        return type(name, bases, namespace)


#
# Serializer class builder
#
def api_info(rest_api, version) -> ApiInfo:
    """
    Builds an ApiInfo object.
    """
    registry_map = rest_api.registries.get(version, {})
    classes = {info.cls for info in registry_map.values()}
    for k, v in rest_api.registries[None].items():
        if v.model not in classes:
            registry_map[k] = v.copy()
    return ApiInfo(version, registry_map)


def as_bases(bases):
    if isinstance(bases, type):
        return (bases,)
    return tuple(bases)


def viewset_meta(info, extra_kwargs):
    """
    Meta class for ViewSet.
    """
    return type('Meta', (), {
        'model': info.model,
        'fields': info.fields,
        'extra_kwargs': extra_kwargs,
    })


def serializer_meta(info, extra_kwargs):
    """
    Meta class for Serializer class.
    """
    return type('Meta', (), {
        'model': info.model,
        'fields': ['links', *info.fields],
        'extra_kwargs': extra_kwargs,
    })



import re
import sys
from collections import deque
from html.parser import HTMLParser

from django.test.client import Client

# Utility functions
cte = (lambda x: lambda *args: x)
choose = (lambda cond, do, other: do if cond else other)


def crawl(url='/', skip=(), errors=(), user=None, log=print):
    """
    Crawl website starting from the given base url and return a dictionary with
    all pages with invalid status codes (e.g. 404, 500, etc)

    Args:
        url (str):
            Starting url.
        skip (list of regex strings):
            List of regular expressions with patterns that should be
            skip even if a hyperlink is found in the webpage.
        errors (list of regex strings):
            List of regular expressions that match links that should be
            considered instant errors.
        user:
            User used to visit the pages.
        log:
            Function used to print debug messages. Uses the builtin print()
            function by default..
    """
    # Create test client
    client = Client()
    if user:
        client.force_login(user)

    # Control urls that should be included/excluded from analysis
    skip_re = re.compile('|'.join(skip))
    errors_re = re.compile('|'.join(errors))
    keep = choose(skip, lambda x: not skip_re.match(x), cte(True))
    is_error = choose(errors, lambda x: errors_re.match(x), cte(False))
    log = log or cte(None)

    # Accumulation variables
    visited = {}
    pending = deque([url])
    errors = {}

    while pending:
        url = pending.popleft()
        if url in visited:
            continue

        response = client.get(url)
        code = response.status_code
        log(f'visited: {url} (code {code})')
        visited[url] = code

        if code == 200:
            text = response.content.decode(response.charset)
            links = find_urls(text, url)
            pending.extend(filter(keep, links))
            errors.update((x, url) for x in links if is_error(x))

        elif code in (301, 302):
            pending.append(response.url)
        else:
            errors[url] = code

    return errors, visited


def check_link_errors(*args, visit=(), user='user', **kwargs):
    """
    Craw site starting from the given base URL and raise an error if the
    resulting error dictionary is not empty.

    Notes:
        Accept the same arguments of the :func:`crawl` function.
    """
    errors, visited = crawl(*args, **kwargs)
    for url in visit:
        if url not in visited:
            errors[url] = f'URL was not visited by {user}'
    if errors:
        for url, code in errors.items():
            if isinstance(code, int):
                print(f'URL {url} returned invalid status code: {code}')
            else:
                print(f'Invalid URL {url} encountered at {code}')
        raise AssertionError(errors, visited)
    return visited


#
# Utility
#
class HTMLAnchorFinder(HTMLParser):
    SKIP_VALUE_RE = re.compile(r'http://|https://|#.*|^$')

    def __init__(self, data=None, current='/'):
        super().__init__()
        if data is None:
            data = set()
        self.data = data
        self.current = current

    def handle_starttag(self, tag, attrs):
        regex = self.SKIP_VALUE_RE

        if tag != 'a':
            return

        for name, href in attrs:
            if name == 'href' and regex.match(href) is None:
                if not href.startswith('/'):
                    href = self.current + href
                self.data.add(href)

    def error(self, message):
        print(message, file=sys.stdout)

    def iter_urls(self):
        return iter(self.data)


def find_urls(src, base_path='/'):
    """
    Find all internal href values in the given source code.

    Normalizes to absolute paths by using the base_url as reference.
    """
    parser = HTMLAnchorFinder(set(), base_path)
    parser.feed(src)
    return parser.iter_urls()



from .value import value


class ValueMap(dict):
    def compute(self, name, *args, **kwargs):
        return name in self and self[name].compute(*args, **kwargs)

    def value_exists(self, name):
        return name in self

    def add_value(self, name, pred):
        if name in self:
            raise KeyError('A rule with name `%s` already exists' % name)
        self[name] = pred

    def remove_value(self, name):
        del self[name]

    def __setitem__(self, name, pred):
        fn = value(pred)
        super().__setitem__(name, fn)


# Shared rule set
default_value_map = ValueMap()
NOT_GIVEN = object()


def add_value(name, func):
    default_value_map.add_value(name, func)


def register_value(name, **kwargs):
    def decorator(func):
        vfunc = value(**kwargs)(func)
        default_value_map.add_value(name, vfunc)
        return vfunc

    return decorator


def remove_value(name):
    default_value_map.remove_value(name)


def value_exists(name):
    return default_value_map.value_exists(name)


def compute(name, *args, **kwargs):
    return default_value_map.compute(name, *args, **kwargs)


def get_value(name, default=NOT_GIVEN):
    try:
        return default_value_map[name]
    except KeyError:
        if default is NOT_GIVEN:
            raise ValueError('could not find value: %s' % name)
        return default



from django.conf import settings


def get_config(key, default=None):
    """
    Get settings from django.conf if exists or return default value otherwise.
    """
    return getattr(settings, key, default)



import importlib
import operator as op

from django.core.exceptions import ImproperlyConfigured
from sidekick import lazy


class LazyMethod:
    """
    Lazily import a method from the given class and inject in the current class
    on first use.
    """

    @lazy
    def object(self):
        try:
            mod = importlib.import_module(self.module)
        except ImportError:
            raise ImproperlyConfigured(
                f'failed to import module: {self.module}\n'
                f'Please install this module in your environment.'
            )
        return self.attr_getter(mod)

    def __init__(self, path, pip=None):
        self.module, _, self.path = path.partition(':')
        self.pip = pip
        self.attr_getter = op.attrgetter(self.path)

    def __get__(self, obj, cls=None):
        imported = self.object
        try:
            getter = imported.__get__
        except AttributeError:
            return self if obj is None else obj
        else:
            return getter(obj, cls)

    def __set__(self, obj, value):
        obj = self.object
        obj.__set__(obj, value)

    def __set_name__(self, owner, name):
        self.name = name


def with_base(bases, cls):
    """
    Return a copy of bases tuple that contains the given additional class `cls`.
    """

    if cls in bases:
        return bases

    bases = list(bases)
    tail = []
    while bases:
        if issubclass(cls, bases[-1]):
            tail.append(bases.pop())
        else:
            bases.append(cls)
            break

    return (*bases, *reversed(tail))



from functools import wraps

from django.db import models
from django.utils.translation import ugettext_lazy as _
from sidekick import import_later

autoslug = import_later('autoslug')


def label(value):
    """
    Add default label to field function.
    """

    def decorator(func):
        @wraps(func)
        def decorated(verbose_name=value, *args, **kwargs):
            return func(verbose_name, *args, **kwargs)

        return decorated

    return decorator


def max_length(name, default=140):
    """
    Reads max_length property from settings.
    """
    return arg_from_settings('max_length', name, default)


def arg_from_settings(prop, name, default=None):
    """
    Read argument from settings file.
    """

    def decorator(func):
        @wraps(func)
        def decorated(*args, **kwargs):
            if prop not in kwargs:
                from django.conf import settings
                kwargs[prop] = getattr(settings, name, default)
            return func(*args, **kwargs)

        return decorated

    return decorator


# noinspection PyPep8Naming
@label(_('slug'))
def AutoSlugField(*args, **kwargs):  # noqa: N802
    return autoslug.AutoSlugField(*args, **kwargs)


# noinspection PyPep8Naming
@label(_('name'))
@max_length('NAME_MAX_LENGTH', 40)
def NameField(*args, **kwargs):  # noqa: N802
    return models.CharField(*args, **kwargs)


# noinspection PyPep8Naming
@label(_('title'))
@max_length('TITLE_MAX_LENGTH', 140)
def TitleField(*args, **kwargs):  # noqa: N802
    return models.CharField(*args, **kwargs)


# noinspection PyPep8Naming
@label(_('description'))
@max_length('SHORT_DESCRIPTION_MAX_LENGTH', 255)
def ShortDescriptionField(*args, **kwargs):  # noqa: N802
    kwargs.setdefault('help_text', _(
        'Conscise description used on listings.'
    ))
    return models.CharField(*args, **kwargs)


# noinspection PyPep8Naming
@label(_('detailed description'))
def LongDescriptionField(*args, **kwargs):  # noqa: N802
    kwargs.setdefault('help_text', _(
        'Detailed description of object.'
    ))
    return models.TextField(*args, **kwargs)



from django.test import Client as DjangoClient
from sidekick import import_later

bs4 = import_later('bs4')
Text = import_later('hyperpython:Text')


class Client(DjangoClient):
    def get_data(self, *args, fix_links=False, **kwargs):
        """
        Makes a get request and return result as a raw string of data instead
        of the usual response object.
        """
        response = self.get(*args, **kwargs)
        if getattr(response, 'url', None):
            return self.get_data(response.url, fix_links=fix_links)

        if fix_links:
            soup = self.get_soup(*args, fix_links=True, **kwargs)
            return str(soup)
        return response.content.decode(response.charset)

    def get_html(self, *args, **kwargs):
        """
        Like .get_data(), but converts the result into an hyperpython Text()
        instance.

        Text instances are rendered as HTML in Jupyter Notebooks.
        """
        return Text(self.get_data(*args, **kwargs), escape=False)

    def get_soup(self, *args, fix_links=False, **kwargs):
        """
        Return response of a request as a Beautiful soup object.
        """
        soup = bs4.BeautifulSoup(self.get_data(*args, **kwargs))
        if fix_links:
            add_href_prefix(soup, 'http://localhost:8000')
        return soup


def add_href_prefix(soup, prefix):
    for link in soup.find_all('a'):
        if link['href'].starswith('/'):
            link['href'] = prefix + link['href']



from collections import Sequence, Mapping
from functools import wraps

from django.db.models import QuerySet
from rules import test_rule
from rules.predicates import NO_VALUE

import sidekick as sk
from sidekick import lazy
from .valuemap import compute


def proxy(obj, user=NO_VALUE, *, values=None, perms=None, rules=None, **kwargs):
    """
    Creates a proxy that augments the given object with all attributes passed
    as keyword arguments. Functions are bound to the obj instance as the
    first argument.
    """
    return Proxy(obj, user, values, perms, rules, kwargs)


def proxy_seq(seq, user=None, *, values=None, perms=None, rules=None, **kwargs):
    """
    Similar to proxy, but expect a sequence/queryset/mapping of objects. It
    augment the items in the collection rather than the collection object
    itself.
    """
    args = (user, values, perms, rules, kwargs)
    if isinstance(seq, QuerySet):
        return QuerySetProxy(seq, *args)
    elif isinstance(seq, Mapping):
        return MappingProxy(seq, *args)
    return CollectionProxy(seq, *args)


#
# Auxiliary classes
#
class Proxy(sk.Proxy):
    """
    Acts as a proxy of objects with arbitrary additional parameters.
    """

    def __init__(self, obj, user, values, perms, rules, kwargs):
        super().__init__(obj)
        self._kwargs = kwargs
        self._user = user
        self._values = values or ()
        self._perms = perms or {}
        self._rules = rules or {}

    def __getattr__(self, attr):
        value = NO_VALUE
        user = self._user
        obj = self._obj__

        if attr in self._kwargs:
            value = self._kwargs[attr]
            if callable(value):
                value = value(self)
        if attr in self._values:
            value_id = self._values[attr]
            value = compute(value_id, obj, user)
        if attr in self._perms and self._user is not NO_VALUE:
            perm = self._perms[attr]
            value = user.has_perm(perm, obj)
        if attr in self._rules:
            rule = self._rules[attr]
            value = test_rule(rule, obj, user)

        if value is NO_VALUE:
            return super().__getattr__(attr)

        setattr(self, attr, value)
        return value


class CollectionProxy(Sequence):
    """
    Acts as a proxy for a immutable collection of objects: i.e., attributes
    are augmented after iteration.
    """
    _args = lazy(lambda x: (x._user, x._values, x._perms, x._rules, x._kwargs))

    def __init__(self, obj, user, values, perms, rules, kwargs):
        self._obj = obj
        self._user = user
        self._values = values
        self._perms = perms
        self._rules = rules
        self._kwargs = kwargs

    def __len__(self):
        return len(self._obj)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return CollectionProxy(self._obj[idx], *self._args)
        return Proxy(self._obj[idx], *self._args)

    def __iter__(self):
        args = self._args
        for obj in self._obj:
            yield Proxy(obj, *args)

    def __bool__(self):
        return bool(self._obj)


class MappingProxy(Mapping, CollectionProxy):
    def __iter__(self):
        return iter(self._obj)

    def items(self):
        args = self._args
        return ((k, Proxy(v, *args)) for k, v in self._obj.items())


class QuerySetProxy(CollectionProxy):
    """
    Acts as a proxy of objects with arbitrary additional parameters.
    """

    def __getattr__(self, attr):
        value = getattr(self._obj, attr)

        if callable(value):
            function = value

            @wraps(function)
            def value(*args, **kwargs):
                result = function(*args, **kwargs)
                if isinstance(result, QuerySet):
                    return QuerySetProxy(result, *self._args)
                return Proxy(result, *self._args)

        setattr(self, attr, value)
        return value



from django.apps import apps
from django.contrib.auth import models as auth
from django.utils.translation import ugettext_lazy as _
from sidekick import property as sk_property, placeholder as this

from boogie import models


class UserQuerySet(models.QuerySet):
    """
    Base queryset for Boogie users.
    """


class UserManager(models.Manager,
                  auth.UserManager.from_queryset(UserQuerySet)):
    """
    Base manager for Boogie users.
    """


class AbstractUser(auth.AbstractUser, models.Model):
    """
    A user object with a single name field instead of separate first_name and
    last_name.

    This is the abstract version of the model. Use it for subclassing.
    """
    name = models.CharField(
        _('Name'),
        max_length=255,
        default='',
        help_text=_('User\'s full name'),
    )

    first_name = sk_property(this.name.partition(' ')[0])
    last_name = sk_property(this.name.partition(' ')[-1])

    @first_name.setter
    def first_name(self, value):
        pre, _, post = self.name.partition(' ')
        self.name = f'{value} {post}' if post else value

    @last_name.setter
    def last_name(self, value):
        pre, _, post = self.name.partition(' ')
        self.name = f'{pre} {value}' if post else value

    objects = UserManager()

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.name:
            self.name = self.username

    def __str__(self):
        return self.email


class User(AbstractUser):
    """
    Concrete version of Boogie's user model.
    """

    class Meta(auth.User.Meta):
        swappable = 'AUTH_USER_MODEL'
        if not apps.is_installed('boogie.apps.users'):
            app_label = 'boogie_users'



from django.db.models import Model
from sidekick import Record, field


class Route(Record):
    """
    Define a route in routes.py

    This is analogous to the url() function in Django.
    """
    route = field(str)
    model = field(Model)
    serializer = field(default=None)
    fields = field(default=None)
    viewset = field(default=None)


def include_router(name):
    """
    Include routes from a routes in your urls.py module.

    This function is used to include routes from a urls.py module and works
    analogously to Django's builtin include() function.
    """



from enum import Enum, IntEnum
from functools import lru_cache

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import ugettext_lazy as _

NoneType = type(None)


class EnumField(models.Field):
    """
    A Field type that wraps an enumeration.

    It represents each field internally as a string or integer.

    Args:
        enum (type):
            A subclass of :class:`enum.Enum`. You should consider using
            :class:`boogie.types.DescriptionEnum` in order to provide
            human-friendly names for each enumeration value.
    """

    description = _('An enumeration field')

    def __init__(self, enum, *args, **kwargs):
        if not (isinstance(enum, type) and issubclass(enum, Enum)):
            raise ImproperlyConfigured(
                'First argument must be a enum.Enum subclass.'
            )
        if not list(enum):
            raise ImproperlyConfigured(
                f'Must be a concrete enumeration. The provided class '
                f'{enum.__qualname__} is empty.'
            )
        if 'choices' in kwargs:
            raise ImproperlyConfigured(
                'Cannot set the choices of an enum field.'
            )
        if not is_integer_enum(enum):
            if 'max_length' not in kwargs:
                kwargs['max_length'] = enum_max_length(enum)
            elif kwargs['max_length'] < enum_max_length(enum):
                raise ImproperlyConfigured(
                    'Maximum length is smaller then the larger enum field '
                    'representation.'
                )

        self.enum = enum
        if is_integer_enum(enum):
            self._impl = models.SmallIntegerField
        else:
            self._impl = models.CharField
        super().__init__(*args, **kwargs)
        self.choices = get_choices_from_enum(enum)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args = [self.enum] + args
        kwargs.pop('choices')
        return name, path, args, kwargs

    def get_internal_type(self):
        return self._impl.get_internal_type(self)

    def to_python(self, value):
        if value in self.empty_values:
            return None
        value = self._impl.to_python(self, value)
        return value_to_enum(self.enum, value)

    def get_db_prep_value(self, value, connection, prepared=False):
        value = getattr(value, 'value', value)
        prep_value = self._impl.get_db_prep_value
        return prep_value(self, value, connection, prepared=prepared)

    def contribute_to_class(self, cls, name, *args, **kwargs):
        super().contribute_to_class(cls, name, *args, **kwargs)

        # Add options
        prefix = name.upper() + '_'
        for option in self.enum:
            setattr(cls, prefix + option.name, option)

        # Create descriptor that wraps field access. The descriptor guarantees
        # that the object is always converted to Enum types
        setattr(cls, name, EnumDescriptor(self.enum, name))

    def formfield(self, **kwargs):
        # FIXME: Super ugly hack! Try to find a more official solution instead
        # of patching a method of a live instance.
        result = super().formfield(**kwargs)
        result.widget.render = fix_renderer(result.widget.render)
        return result


class EnumDescriptor:
    def __init__(self, enum, name):
        self.enum = enum
        self.name = name

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        value = instance.__dict__.get(self.name)
        enum = self.enum
        if isinstance(value, (enum, NoneType)):
            return value
        value = value_to_enum(enum, value)
        instance.__dict__[self.name] = value
        return value

    def __set__(self, instance, value):
        enum = self.enum
        if value == '':
            value = next(iter(enum))
        if not isinstance(value, (enum, NoneType)):
            value = value_to_enum(enum, value)
        instance.__dict__[self.name] = value


def fix_renderer(renderer):
    """
    Patch the .render() function of a select widget to use the .value of a enum
    field instead of a Enum instance.
    """

    def render(value=None, **kwargs):
        value = getattr(value, 'value', value)
        return renderer(value=value, **kwargs)

    return render


def get_choices_from_enum(enum):
    """
    Return a list of (name, verbose name) choices from an Enum type.
    """

    def human(x):
        return x.lower().replace('_', ' ')

    def description(x):
        try:
            return x.description
        except AttributeError:
            return human(x.name)

    if is_integer_enum(enum):
        return tuple((e.value, description(e)) for e in enum)
    else:
        return tuple((e.value, human(e.name)) for e in enum)


@lru_cache(2048)  # noqa C901
def value_to_enum(enum_type, value):
    """
    Create Enum instance from a string value. This will scan the list of
    enumerations if string is not found.
    """

    # Special case valid values
    if isinstance(value, (enum_type, NoneType)):
        return value

    # Simple transformation to enum
    try:
        return enum_type(value)
    except ValueError:
        pass

    # Some types do not
    for obj in enum_type:
        if obj.value == value:
            return obj

    # Maybe we provided the enum name
    if isinstance(value, str):
        try:
            new_value = getattr(enum_type, value)
            if isinstance(new_value, enum_type):
                return new_value
        except (AttributeError, TypeError):
            pass

        # Sometimes the string comes in the form of <TypeName>.<Enum Name>
        if value.startswith(enum_type.__name__ + '.'):
            attr = value[len(enum_type.__name__) + 1:]
            if attr and hasattr(enum_type, attr):
                new_value = getattr(enum_type, attr)
                if isinstance(new_value, enum_type):
                    return new_value

    # Check if value can be coerced to string
    try:
        return enum_type(str(value))
    except ValueError:
        pass

    # Give up!
    raise ValueError('not a valid value for enumeration: %r' % value)


@lru_cache(256)
def is_integer_enum(tt):
    if isinstance(tt, IntEnum):
        return True
    else:
        return all(isinstance(opt.value, int) for opt in tt)


def enum_max_length(tt):
    return max(len(str(opt.value)) for opt in tt)



from pathlib import Path

from django.core.exceptions import ImproperlyConfigured

from .paths import PathsConf


class LoggingConf(PathsConf):
    """
    Base configurations for logging.
    """

    def get_debug_logger(self):
        return {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        }

    def get_default_logger(self):
        return {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        }

    def get_logging_console_handler(self):
        return {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        }

    def get_logging_file_handler(self):
        path = Path(self.LOG_FILE_PATH)
        if path.is_dir():
            raise ImproperlyConfigured(
                'log file is a directory. Please set a correct LOG_FILE_PATH '
                'in your configuration file that points to a log file, instead '
                'of a directory.'
            )
        return {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': path,
        }

    def get_logging(self):  # noqa: N802
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'handlers': {
                'file': self.LOGGING_CONSOLE_HANDLER,
                'console': self.LOGGING_FILE_HANDLER,
            },
            'loggers': {
                'ej': self.DEBUG_LOGGER,
                'django': self.DEFAULT_LOGGER,
                'celery': self.DEFAULT_LOGGER,
            },
        }



import traceback
from collections import OrderedDict

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from rest_framework import serializers
from rest_framework.relations import HyperlinkedRelatedField
from rest_framework.serializers import raise_errors_on_nested_writes
from rest_framework.utils import model_meta
from sidekick import lazy

from .settings import get_url_prefix
from .utils import join_url


class RestAPIRelatedField(HyperlinkedRelatedField):
    def to_internal_value(self, data):
        return self.get_queryset().get(**{self.lookup_field: data})

    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict(
            (getattr(item, self.lookup_field), self.display_value(item))
            for item in queryset
        )

    def to_representation(self, value):
        return str(value)

    def use_pk_only_optimization(self):
        return False


class WithLinksSerializerMixin(serializers.Serializer):
    serializer_related_field = RestAPIRelatedField
    links = serializers.SerializerMethodField()
    api_version = None

    # Actions and extra links
    explicit_links = ()

    @lazy
    def request(self):
        return self.context['request']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Define base urls
        request = self.request
        if request is None:
            self.url_prefix = ''
        else:
            self.url_prefix = get_url_prefix(request)

    def get_links(self, obj):
        """
        Return the links dictionary mapping resource names to their
        corresponding links according to HATEAOS.
        """
        return {
            **self._inner_links(obj),
            **self._explicit_links(obj),
            **self._outer_links(obj),
        }

    def _inner_links(self, obj):
        return {}

    def _explicit_links(self, obj):
        """
        Return a mapping with all explicitly registered hyperlinks.
        """
        request = self.request
        return {name: func(request, obj) for name, func in self.explicit_links}

    def _outer_links(self, obj):
        """
        Return a mapping with all external hyperlinks following from the given
        object.
        """
        links = {}

        for name, params in self.Meta.extra_kwargs.items():
            value = getattr(obj, name)
            lookup_field = params['lookup_field']
            lookup_value = getattr(value, lookup_field)

            kwargs = {lookup_field: lookup_value}
            view_name = params['view_name']
            url = reverse(view_name, kwargs=kwargs)

            links[name] = self.url_prefix + url

        return links


class RestAPISerializer(WithLinksSerializerMixin, serializers.ModelSerializer):
    """
    An extended HyperlinkedModelSerializer:

    * Puts all relations to external resources on the obj.links attributes
      (including a self link)
    * Foreign key relations produce links on the links attribute
    """

    # Urls names and lookup
    base_name = None
    detail_url = None
    list_url = None
    lookup_field = None

    # Actions and extra links
    actions = ()
    explicit_links = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.base_name is None:
            cls = type(self)
            name = '%s.%s' % (cls.__module__, cls.__qualname__)
            raise ImproperlyConfigured(
                'Must provide a "base_name" value for the RestAPISerializer '
                'subclass. %s does not define such class attribute.' % name
            )
        base_path = reverse(self.base_name + '-list')
        self.base_url = join_url(self.url_prefix, base_path)

    def _inner_links(self, obj):
        """
        Return a mapping of all inner hyperlinks in object.
        """

        lookup_field = self.lookup_field
        kwargs = {lookup_field: getattr(obj, lookup_field)}
        self_url = reverse(self.detail_url, kwargs=kwargs)
        self_url = self.url_prefix + self_url

        extra = {action: f'{self_url}{action}/' for action in self.actions}
        return {'self': self_url, **extra}

    #
    # Overloads serializer methods
    #
    def create(self, validated_data):
        # We cannot call the super method because DRF explicitly calls
        # manager.create() inside the method. This is essentially the
        # same method in ModelSerializer, but instead of creating the object
        # directly, it uses the save_hook method.
        raise_errors_on_nested_writes('create', self, validated_data)

        model_class = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved.
        info = model_meta.get_field_info(model_class)
        many_to_many = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        try:
            instance = model_class(**validated_data)
        except TypeError:
            tb = traceback.format_exc()
            msg = ('Got a `TypeError` when saving object. '
                   'This may be because you have a writable field on the '
                   'serializer class that is not a valid argument to the '
                   'object constructor. You may need to make the field '
                   'read-only, or register a save hook function to handle '
                   'this correctly.\nOriginal exception was:\n %s' % tb)
            raise TypeError(msg)
        else:
            instance = self.save_hook(self.request, instance)

        # Save many-to-many relationships after the instance is created.
        if many_to_many:
            for field_name, value in many_to_many.items():
                field = getattr(instance, field_name)
                field.set(value)

        return instance

    def update(self, instance, validated_data):
        raise_errors_on_nested_writes('update', self, validated_data)
        info = model_meta.get_field_info(instance)

        # Simply set each attribute on the instance, and then save it.
        # Note that unlike `.create()` we don't need to treat many-to-many
        # relationships as being a special case. During updates we already
        # have an instance pk for the relationships to be associated with.
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)

        return self.save_hook(self.request, instance)

    #
    # Hooks
    #
    def save_hook(self, request, instance):
        instance.save()
        return instance


class RestAPIInlineSerializer(WithLinksSerializerMixin,
                              serializers.ModelSerializer):
    """
    Base serializer class for inline models.
    """



from functools import wraps

from celery import shared_task
from pushtogether_math import cluster

from ej_conversations.models import Conversation
from .models import Job


# decorator to avoid code duplication
def update_job(fn):
    """
    Decorator that will update Job with result of the function
    """

    # wraps will make the name and docstring of fn available for introspection
    @wraps(fn)
    def wrapper(job_id, *args, **kwargs):
        job = Job.objects.get(pk=job_id)
        job.status = Job.STARTED
        job.save()
        try:
            # execute the function fn
            result = fn(*args, **kwargs)
            job.result = result
            job.status = Job.FINISHED if result else Job.STUCK
            job.save()
        except:
            job.result = None
            job.status = Job.FAILED
        job.save()

    return wrapper


@shared_task
@update_job
def get_clusters(conversation_id):
    """
    Celery task that receives an conversation ID, checks if this conversation
    has sufficient data to be clustered, than returns the clusters if it's ok,
    or None if it isn't.
    """
    conversation = Conversation.objects.get(pk=conversation_id)
    if has_sufficient_data(conversation):
        votes = conversation.get_vote_data()
        return cluster.get_clusters(votes, range(2, 5))


def has_sufficient_data(conversation):
    """
    Check if the conversation available data respect the limits of the
    math configuration variables.
    """
    if (conversation.total_participants < Job.MATH_MIN_USERS or
        conversation.total_comments < Job.MATH_MIN_COMMENTS or
        conversation.total_votes < Job.MATH_MIN_VOTES):
        return False
    return True


TASK_MAPPING = {
    'CLUSTERS': get_clusters,
}



from django.db.models import functions

from .combinable import get_comparable_class

concat = get_comparable_class(functions.Concat, """
Concatenates the values of all given fields or expressions.

Usage:
    concat(F.first_name, ' ', F.last_name)
""")

coalesce = get_comparable_class(functions.Coalesce, """
Chooses the first non-null value from left to right.

Usage:
    coalesce(F.savings_account, F.check_account, 0)
""")

greatest = get_comparable_class(functions.Greatest, """
Chooses the greatest of the given arguments.

Usage:
    greatest(F.savings_account, F.check_account, 0)
""")

least = get_comparable_class(functions.Least, """
Chooses the smallest of the given arguments.

Usage:
    least(F.savings_account, F.check_account, 0)
""")

length = get_comparable_class(functions.Length, """
Returns the length of a string.

Usage:
    length(F.name)
""")
